"""Model for OracleSet transaction type."""

from __future__ import annotations

import datetime
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

MAX_ORACLE_DATA_SERIES = 10
MAX_ORACLE_PROVIDER = 256
MAX_ORACLE_URI = 256
MAX_ORACLE_SYMBOL_CLASS = 16

# epoch offset must equal 946684800 seconds. It represents the diff between the
# genesis of Unix time and ripple epoch time
EPOCH_OFFSET = int(
    (datetime.datetime(2000, 1, 1) - datetime.datetime(1970, 1, 1)).total_seconds()
)


@require_kwargs_on_init
@dataclass(frozen=True)
class OracleSet(Transaction):
    """Creates a new Oracle ledger entry or updates the fields of an existing one,
    using the Oracle ID.

    The oracle provider must complete these steps before submitting this transaction:

    Create or own the XRPL account in the Owner field and have enough XRP to meet the
    reserve and transaction fee requirements.
    Publish the XRPL account public key, so it can be used for verification by dApps.
    Publish a registry of available price oracles with their unique OracleDocumentID .
    """

    account: str = REQUIRED  # type: ignore
    """This account must match the account in the Owner field of the Oracle object."""

    oracle_document_id: int = REQUIRED  # type: ignore
    """A unique identifier of the price oracle for the Account."""

    provider: Optional[str] = None
    """
    This field must be hex-encoded. You can use `xrpl.utils.str_to_hex` to
    convert a UTF-8 string to hex.

    An arbitrary value that identifies an oracle provider, such as Chainlink, Band, or
    DIA. This field is a string, up to 256 ASCII hex encoded characters (0x20-0x7E).
    This field is required when creating a new Oracle ledger entry, but is optional for
    updates.
    """

    uri: Optional[str] = None
    """
    This field must be hex-encoded. You can use `xrpl.utils.str_to_hex` to
    convert a UTF-8 string to hex.

    An optional Universal Resource Identifier to reference price data off-chain. This
    field is limited to 256 bytes.
    """

    asset_class: Optional[str] = None
    """
    This field must be hex-encoded. You can use `xrpl.utils.str_to_hex` to
    convert a UTF-8 string to hex.

    Describes the type of asset, such as "currency", "commodity", or "index". This
    field is a string, up to 16 ASCII hex encoded characters (0x20-0x7E). This field is
    required when creating a new Oracle ledger entry, but is optional for updates.
    """

    last_update_time: int = REQUIRED  # type: ignore
    """LastUpdateTime is the specific point in time when the data was last updated.
    The LastUpdateTime is represented as Unix Time - the number of seconds since
    January 1, 1970 (00:00 UTC)."""

    price_data_series: List[PriceData] = REQUIRED  # type: ignore
    """An array of up to 10 PriceData objects, each representing the price information
    for a token pair. More than five PriceData objects require two owner reserves."""

    transaction_type: TransactionType = field(
        default=TransactionType.ORACLE_SET,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        # If price_data_series is not set, do not perform further validation
        if "price_data_series" not in errors:
            if len(self.price_data_series) == 0:
                errors["price_data_series"] = "Field must have a length greater than 0."

            if len(self.price_data_series) > MAX_ORACLE_DATA_SERIES:
                errors["price_data_series"] = (
                    "Field must have a length less than"
                    f" or equal to {MAX_ORACLE_DATA_SERIES}"
                )

            # either asset_price and scale are both present or both excluded
            for price_data in self.price_data_series:
                if (price_data.asset_price is not None) != (
                    price_data.scale is not None
                ):
                    errors["price_data_series"] = (
                        "Field must have both "
                        "`AssetPrice` and `Scale` if any are present"
                    )

        if self.asset_class is not None and len(self.asset_class) == 0:
            errors["asset_class"] = "Field must have a length greater than 0."

        if (
            self.asset_class is not None
            and len(self.asset_class) > MAX_ORACLE_SYMBOL_CLASS
        ):
            errors["asset_class"] = (
                "Field must have a length less than"
                f" or equal to {MAX_ORACLE_SYMBOL_CLASS}"
            )

        if self.provider is not None and len(self.provider) == 0:
            errors["provider"] = "Field must have a length greater than 0."

        if self.provider is not None and len(self.provider) > MAX_ORACLE_PROVIDER:
            errors["provider"] = (
                f"Field must have a length less than or equal to {MAX_ORACLE_PROVIDER}."
            )

        if self.uri is not None and len(self.uri) == 0:
            errors["uri"] = "Field must have a length greater than 0."

        if self.uri is not None and len(self.uri) > MAX_ORACLE_URI:
            errors["uri"] = (
                f"Field must have a length less than or equal to {MAX_ORACLE_URI}."
            )

        # check on the last_update_time
        if self.last_update_time < EPOCH_OFFSET:
            errors["last_update_time"] = (
                "LastUpdateTime must be greater than or equal"
                f" to ripple epoch - {EPOCH_OFFSET} seconds"
            )

        return errors


@require_kwargs_on_init
@dataclass(frozen=True)
class PriceData(NestedModel):
    """Represents one PriceData element. It is used in OracleSet transaction"""

    base_asset: str = REQUIRED  # type: ignore
    """The primary asset in a trading pair. Any valid identifier, such as a stock
    symbol, bond CUSIP, or currency code is allowed. For example, in the BTC/USD pair,
    BTC is the base asset; in 912810RR9/BTC, 912810RR9 is the base asset."""

    quote_asset: str = REQUIRED  # type: ignore
    """The quote asset in a trading pair. The quote asset denotes the price of one unit
    of the base asset. For example, in the BTC/USD pair, BTC is the base asset; in
    912810RR9/BTC, 912810RR9 is the base asset."""

    asset_price: Optional[int] = None
    """The asset price after applying the Scale precision level. It's not included if
    the last update transaction didn't include the BaseAsset/QuoteAsset pair."""

    scale: Optional[int] = None
    """The scaling factor to apply to an asset price. For example, if Scale is 6 and
    original price is 0.155, then the scaled price is 155000. Valid scale ranges are
    0-10. It's not included if the last update transaction didn't include the
    BaseAsset/QuoteAsset pair."""



"""Model for a XChainAddAccountCreateAttestation transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union

from typing_extensions import Literal

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainAddAccountCreateAttestation(Transaction):
    """
    Represents a XChainAddAccountCreateAttestation transaction.
    The XChainAddAccountCreateAttestation transaction provides an attestation
    from a witness server that a XChainAccountCreateCommit transaction occurred
    on the other chain.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge associated with the attestation. This field is required.

    :meta hide-value:
    """

    public_key: str = REQUIRED  # type: ignore
    """
    The public key used to verify the signature. This field is required.

    :meta hide-value:
    """

    signature: str = REQUIRED  # type: ignore
    """
    The signature attesting to the event on the other chain. This field is
    required.

    :meta hide-value:
    """

    other_chain_source: str = REQUIRED  # type: ignore
    """
    The account on the source chain that submitted the
    ``XChainAccountCreateCommit`` transaction that triggered the event
    associated with the attestation. This field is required.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The amount committed by the ``XChainAccountCreateCommit`` transaction on
    the source chain. This field is required.

    :meta hide-value:
    """

    attestation_reward_account: str = REQUIRED  # type: ignore
    """
    The account that should receive this signer's share of the
    ``SignatureReward``. This field is required.

    :meta hide-value:
    """

    attestation_signer_account: str = REQUIRED  # type: ignore
    """
    The account on the door account's signer list that is signing the
    transaction. This field is required.

    :meta hide-value:
    """

    was_locking_chain_send: Union[Literal[0], Literal[1]] = REQUIRED  # type: ignore
    """
    A boolean representing the chain where the event occurred. This field is
    required.

    :meta hide-value:
    """

    xchain_account_create_count: Union[str, int] = REQUIRED  # type: ignore
    """
    The counter that represents the order that the claims must be processed in.
    This field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The destination account for the funds on the destination chain. This field
    is required.

    :meta hide-value:
    """

    signature_reward: Amount = REQUIRED  # type: ignore
    """
    The signature reward paid in the ``XChainAccountCreateCommit`` transaction.
    This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_ADD_ACCOUNT_CREATE_ATTESTATION,
        init=False,
    )



"""Model for a XChainCreateBridge transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.currencies import XRP
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainCreateBridge(Transaction):
    """
    Represents a XChainCreateBridge transaction.
    The XChainCreateBridge transaction creates a new `Bridge` ledger object and
    defines a new cross-chain bridge entrance on the chain that the transaction
    is submitted on. It includes information about door accounts and assets for
    the bridge.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge (door accounts and assets) to create. This field is required.

    :meta hide-value:
    """

    signature_reward: str = REQUIRED  # type: ignore
    """
    The total amount to pay the witness servers for their signatures. This
    amount will be split among the signers. This field is required.

    :meta hide-value:
    """

    min_account_create_amount: Optional[str] = None
    """
    The minimum amount, in XRP, required for a ``XChainAccountCreateCommit``
    transaction. If this isn't present, the ``XChainAccountCreateCommit``
    transaction will fail. This field can only be present on XRP-XRP bridges.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_CREATE_BRIDGE,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge

        if bridge.locking_chain_door == bridge.issuing_chain_door:
            errors["xchain_bridge"] = (
                "Cannot have the same door accounts on the locking and issuing chain."
            )

        if self.account not in [bridge.locking_chain_door, bridge.issuing_chain_door]:
            errors["account"] = (
                "account must be either locking chain door or issuing chain door."
            )

        if (bridge.locking_chain_issue == XRP()) != (
            bridge.issuing_chain_issue == XRP()
        ):
            errors["issue"] = "Bridge must be XRP-XRP or IOU-IOU."

        if (
            self.min_account_create_amount is not None
            and bridge.locking_chain_issue != XRP()
        ):
            errors["min_account_create_amount"] = (
                "Cannot have MinAccountCreateAmount if bridge is IOU-IOU."
            )

        if self.signature_reward != REQUIRED and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "signature_reward must be numeric."

        if (
            self.min_account_create_amount is not None
            and not self.min_account_create_amount.isnumeric()
        ):
            errors["min_account_create_amount_value"] = (
                "min_account_create_amount must be numeric."
            )

        return errors



"""Model for Payment transaction type and related flags."""

from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.amounts import Amount, is_xrp
from xrpl.models.path import Path
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    KW_ONLY_DATACLASS,
    require_kwargs_on_init,
    validate_credential_ids,
)


class PaymentFlag(int, Enum):
    """
    Transactions of the Payment type support additional values in the Flags field.
    This enum represents those options.

    `See Payment Flags <https://xrpl.org/payment.html#payment-flags>`_
    """

    TF_NO_RIPPLE_DIRECT = 0x00010000
    """
    Do not use the default path; only use paths included in the Paths field.
    This is intended to force the transaction to take arbitrage opportunities.
    Most clients do not need this.
    """

    TF_PARTIAL_PAYMENT = 0x00020000
    """
    If the specified Amount cannot be sent without spending more than SendMax,
    reduce the received amount instead of failing outright.
    See `Partial Payments <https://xrpl.org/partial-payments.html>`_ for more details.
    """

    TF_LIMIT_QUALITY = 0x00040000
    """
    Only take paths where all the conversions have an input:output ratio
    that is equal or better than the ratio of Amount:SendMax.
    See `Limit <https://xrpl.org/payment.html#limit-quality>`_ Quality for details.
    """


class PaymentFlagInterface(TransactionFlagInterface):
    """
    Transactions of the Payment type support additional values in the Flags field.
    This TypedDict represents those options.

    `See Payment Flags <https://xrpl.org/payment.html#payment-flags>`_
    """

    TF_NO_RIPPLE_DIRECT: bool
    TF_PARTIAL_PAYMENT: bool
    TF_LIMIT_QUALITY: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Payment(Transaction):
    """
    Represents a Payment <https://xrpl.org/payment.html>`_ transaction, which
    sends value from one account to another. (Depending on the path taken, this
    can involve additional exchanges of value, which occur atomically.) This
    transaction type can be used for several `types of payments
    <http://xrpl.local/payment.html#types-of-payments>`_.

    Payments are also the only way to `create accounts
    <http://xrpl.local/payment.html#creating-accounts>`_.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The amount of currency to deliver. If the Partial Payment flag is set,
    deliver *up to* this amount instead. This field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The address of the account receiving the payment. This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Payment, or a hosted recipient to pay.
    """

    invoice_id: Optional[str] = None  # TODO: should be a 256 bit hash
    """
    Arbitrary 256-bit hash representing a specific reason or identifier for
    this Check.
    """

    paths: Optional[List[Path]] = None
    """
    Array of payment paths to be used (for a cross-currency payment). Must be
    omitted for XRP-to-XRP transactions.
    """

    send_max: Optional[Amount] = None
    """
    Maximum amount of source currency this transaction is allowed to cost,
    including `transfer fees <http://xrpl.local/transfer-fees.html>`_,
    exchange rates, and slippage. Does not include the XRP destroyed as a
    cost for submitting the transaction. Must be supplied for cross-currency
    or cross-issue payments. Must be omitted for XRP-to-XRP payments.
    """

    deliver_min: Optional[Amount] = None
    """
    Minimum amount of destination currency this transaction should deliver.
    Only valid if this is a partial payment. If omitted, any positive amount
    is considered a success.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.PAYMENT,
        init=False,
    )

    credential_ids: Optional[List[str]] = None
    """Credentials associated with sender of this transaction. The credentials included
    must not be expired."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        errors.update(validate_credential_ids(self.credential_ids))

        # XRP transaction errors
        if is_xrp(self.amount) and self.send_max is None:
            if self.paths is not None:
                errors["paths"] = "An XRP-to-XRP payment cannot contain paths."
            if self.account == self.destination:
                errors["destination"] = (
                    "An XRP payment transaction cannot have the same sender and "
                    "destination."
                )

        # partial payment errors
        elif self.has_flag(PaymentFlag.TF_PARTIAL_PAYMENT) and self.send_max is None:
            errors["send_max"] = "A partial payment must have a `send_max` value."
        elif self.deliver_min is not None and not self.has_flag(
            PaymentFlag.TF_PARTIAL_PAYMENT
        ):
            errors["deliver_min"] = (
                "A non-partial payment cannot have a `deliver_min` field."
            )

        elif (
            is_xrp(self.amount)
            and (self.send_max and is_xrp(self.send_max))
            and not self.has_flag(PaymentFlag.TF_PARTIAL_PAYMENT)
        ):
            errors["send_max"] = (
                "A non-partial payment cannot have both ``amount`` and `send_max` be "
                "XRP."
            )

        # currency conversion errors
        elif self.account == self.destination:
            if self.send_max is None:
                errors["send_max"] = (
                    "A currency conversion requires a `send_max` value."
                )

        return errors



"""High-level ledger methods with the XRPL ledger."""

from typing import Optional, cast

from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.asyncio.ledger.utils import calculate_fee_dynamically
from xrpl.constants import XRPLException
from xrpl.models.requests import Fee, Ledger
from xrpl.utils import xrp_to_drops


async def get_latest_validated_ledger_sequence(client: Client) -> int:
    """
    Returns the sequence number of the latest validated ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest validated ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Ledger(ledger_index="validated"))
    if response.is_successful():
        return cast(int, response.result["ledger_index"])

    raise XRPLRequestFailureException(response.result)


async def get_latest_open_ledger_sequence(client: Client) -> int:
    """
    Returns the sequence number of the latest open ledger.

    Note: While specifying ledgers in the Ledger RPC call, the appropriate ledger_index
    is "current", not "open". However, this method signature is retained as-is for
    backwards compatibility.

    Read this document for more info:
    https://xrpl.org/docs/references/http-websocket-apis/public-api-methods/ledger-methods/ledger#request-format

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest open ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Ledger(ledger_index="current"))

    if response.is_successful():
        # validate that the `current` ledger is retrieved
        if response.result["ledger"]["closed"] is True:
            raise XRPLException("`closed` key found while requesting current ledger")
        if response.result["validated"] is True:
            raise XRPLException("`validated` key found while requesting current ledger")
        if "ledger_index" in response.result:
            raise XRPLException("`ledger_index` key is not relevant in current ledger")
        if "ledger_current_index" not in response.result:
            raise XRPLException(
                "`ledger_current_index` key must be present in current ledger"
            )

        return cast(int, response.result["ledger_current_index"])

    raise XRPLRequestFailureException(response.result)


async def get_fee(
    client: Client, *, max_fee: Optional[float] = 2, fee_type: str = "open"
) -> str:
    """
    Query the ledger for the current transaction fee.

    Args:
        client: the network client used to make network calls.
        max_fee: The maximum fee in XRP that the user wants to pay. If load gets too
            high, then the fees will not scale past the maximum fee. If None, there is
            no ceiling for the fee. The default is 2 XRP.
        fee_type: The type of fee to return. The options are "open" (the load-scaled
            fee to get into the open ledger), "minimum" (the minimum transaction
            fee) or "dynamic" (dynamic fee-calculation based on the queue size
            of the node). The default is "open". The recommended option is
            "dynamic".

    Returns:
        The transaction fee, in drops.
        `Read more about drops <https://xrpl.org/currency-formats.html#xrp-amounts>`_

    Raises:
        XRPLException: if an incorrect option for `fee_type` is passed in.
        XRPLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Fee())
    if not response.is_successful():
        raise XRPLRequestFailureException(response.result)

    result = response.result
    drops = result["drops"]
    if fee_type == "open":
        fee = cast(str, drops["open_ledger_fee"])
    elif fee_type == "minimum":
        fee = cast(str, drops["minimum_fee"])
    elif fee_type == "dynamic":
        fee = calculate_fee_dynamically(fee_data_set=result)
    else:
        raise XRPLException(
            '`fee_type` param must be "open", "minimum" or "dynamic".'
            f" {fee_type} is not a valid option."
        )
    if max_fee is not None:
        max_fee_drops = int(xrp_to_drops(max_fee))
        if max_fee_drops < int(fee):
            fee = str(max_fee_drops)
    return fee



"""
The fee command reports the current state of the open-ledger requirements
for the transaction cost. This requires the FeeEscalation amendment to be
enabled.

This is a public command available to unprivileged users.
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Fee(Request):
    """
    The fee command reports the current state of the open-ledger requirements
    for the transaction cost. This requires the FeeEscalation amendment to be
    enabled.

    This is a public command available to unprivileged users.
    """

    method: RequestMethod = field(default=RequestMethod.FEE, init=False)



"""Helper functions to normalize an affected node."""

from typing import List, Optional, Union, cast

from typing_extensions import Literal, TypedDict

from xrpl.models import TransactionMetadata
from xrpl.models.transactions.metadata import (
    CreatedNode,
    CreatedNodeFields,
    DeletedNode,
    DeletedNodeFields,
    Fields,
    ModifiedNode,
    ModifiedNodeFields,
)


class NormalizedNode(TypedDict):
    """A model representing an affected node in a standard format."""

    NodeType: Literal["CreatedNode", "ModifiedNode", "DeletedNode"]
    LedgerEntryType: str
    LedgerIndex: str
    NewFields: Optional[Fields]
    FinalFields: Optional[Fields]
    PreviousFields: Optional[Fields]
    PreviousTxnID: Optional[str]
    PreviousTxnLgrSeq: Optional[int]


def _normalize_node(
    affected_node: Union[CreatedNode, ModifiedNode, DeletedNode]
) -> NormalizedNode:
    node_keys = affected_node.keys()
    assert len(node_keys) == 1
    diff_type = cast(
        Literal["CreatedNode", "ModifiedNode", "DeletedNode"],
        list(node_keys)[0],
    )
    if diff_type == "CreatedNode":
        node: Union[CreatedNodeFields, ModifiedNodeFields, DeletedNodeFields] = cast(
            CreatedNode, affected_node
        )["CreatedNode"]
    elif diff_type == "ModifiedNode":
        node = cast(ModifiedNode, affected_node)["ModifiedNode"]
    else:
        node = cast(DeletedNode, affected_node)["DeletedNode"]
    ledger_entry_type = node["LedgerEntryType"]
    ledger_index = node["LedgerIndex"]
    new_fields = cast(Optional[Fields], node.get("NewFields"))
    previous_fields = cast(Optional[Fields], node.get("PreviousFields"))
    final_fields = cast(Optional[Fields], node.get("FinalFields"))
    previous_txn_id = cast(Optional[str], node.get("PreviousTxnID"))
    previous_txn_lgr_seq = cast(Optional[int], node.get("PreviousTxnLgrSeq"))
    return NormalizedNode(
        NodeType=diff_type,
        LedgerEntryType=ledger_entry_type,
        LedgerIndex=ledger_index,
        NewFields=new_fields,
        PreviousFields=previous_fields,
        FinalFields=final_fields,
        PreviousTxnID=previous_txn_id,
        PreviousTxnLgrSeq=previous_txn_lgr_seq,
    )


def normalize_nodes(metadata: TransactionMetadata) -> List[NormalizedNode]:
    """
    Normalize all nodes of a transaction's metadata.

    Args:
        metadata: The transaction's metadata.

    Returns:
        The normalized nodes.
    """
    return [_normalize_node(node) for node in metadata["AffectedNodes"]]



"""Model for SetRegularKey transaction type."""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class SetRegularKey(Transaction):
    """
    Represents a `SetRegularKey <https://xrpl.org/setregularkey.html>`_
    transaction, which assigns, changes, or removes a secondary "regular" key pair
    associated with an account.
    """

    regular_key: Optional[str] = None
    """
    The classic address derived from the key pair to authorize for this
    account. If omitted, removes any existing regular key pair from the
    account. Must not match the account's master key pair.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.SET_REGULAR_KEY,
        init=False,
    )



"""Codec for serializing and deserializing vectors of Hash256."""

from __future__ import annotations

from typing import List, Optional, Type

from typing_extensions import Final, Self

from xrpl.core.binarycodec import XRPLBinaryCodecException
from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.types.hash256 import Hash256
from xrpl.core.binarycodec.types.serialized_type import SerializedType

_HASH_LENGTH_BYTES: Final[int] = 32


class Vector256(SerializedType):
    """Codec for serializing and deserializing vectors of Hash256."""

    def __init__(self: Self, buffer: bytes) -> None:
        """Construct a Vector256."""
        super().__init__(buffer)

    @classmethod
    def from_value(cls: Type[Self], value: List[str]) -> Self:
        """Construct a Vector256 from a list of strings.

        Args:
            value: A list of hashes encoded as hex strings.

        Returns:
            A Vector256 object representing these hashes.

        Raises:
            XRPLBinaryCodecException: If the supplied value is of the wrong type.
        """
        if not isinstance(value, list):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a Vector256: expected list,"
                " received {value.__class__.__name__}."
            )

        byte_list = []
        for string in value:
            byte_list.append(bytes(Hash256.from_value(string)))
        return cls(b"".join(byte_list))

    @classmethod
    def from_parser(
        cls: Type[Self], parser: BinaryParser, length_hint: Optional[int] = None
    ) -> Self:
        """Construct a Vector256 from a BinaryParser.

        Args:
            parser: The parser to construct a Vector256 from.
            length_hint: The number of bytes to consume from the parser.

        Returns:
            A Vector256 object.
        """
        byte_list = []
        num_bytes = length_hint if length_hint is not None else len(parser)
        num_hashes = num_bytes // _HASH_LENGTH_BYTES
        for i in range(num_hashes):
            byte_list.append(bytes(Hash256.from_parser(parser)))
        return cls(b"".join(byte_list))

    def to_json(self: Self) -> List[str]:
        """Return a list of hashes encoded as hex strings.

        Returns:
            The JSON representation of this Vector256.

        Raises:
            XRPLBinaryCodecException: If the number of bytes in the buffer
                                        is not a multiple of the hash length.
        """
        if len(self.buffer) % _HASH_LENGTH_BYTES != 0:
            raise XRPLBinaryCodecException("Invalid bytes for Vector256.")
        hash_list = []
        for i in range(0, len(self.buffer), _HASH_LENGTH_BYTES):
            hash_list.append(self.buffer[i : i + _HASH_LENGTH_BYTES].hex().upper())
        return hash_list



"""Model for EscrowCancel transaction type."""

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class EscrowCancel(Transaction):
    """
    Represents an `EscrowCancel <https://xrpl.org/escrowcancel.html>`_
    transaction, which returns escrowed XRP to the sender after the Escrow has
    expired.
    """

    owner: str = REQUIRED  # type: ignore
    """
    The address of the account that funded the Escrow. This field is required.

    :meta hide-value:
    """

    offer_sequence: int = REQUIRED  # type: ignore
    """
    Transaction sequence (or Ticket number) of the EscrowCreate transaction
    that created the Escrow. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ESCROW_CANCEL,
        init=False,
    )



"""Model for SetFee pseudo-transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.transactions.pseudo_transactions.pseudo_transaction import (
    PseudoTransaction,
)
from xrpl.models.transactions.types import PseudoTransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class SetFee(PseudoTransaction):
    """
    A SetFee pseudo-transaction marks a change in `transaction cost
    <https://xrpl.org/transaction-cost.html>`_ or `reserve requirements
    <https://xrpl.org/reserves.html>`_ as a result of `Fee Voting
    <https://xrpl.org/fee-voting.html>`_.

    The parameters are different depending on if this is before or after the
    `XRPFees Amendment<https://xrpl.org/known-amendments.html#xrpfees>`_

    Before the XRPFees Amendment which was proposed in rippled 1.10.0
    base_fee, reference_fee_units, reserve_base, and reserve_increment
    were required fields.

    After the XRPFees Amendment, base_fee_drops, reserve_base_drops,
    and reserve_increment_drops are required fields.

    No SetFee Pseudo Transaction should contain fields from BOTH before
    and after the XRPFees amendment.
    """

    # Required BEFORE the XRPFees Amendment

    base_fee: Optional[str] = None
    """
    The charge, in drops of XRP, for the reference transaction, as hex. (This is the
    transaction cost before scaling for load.) This field is required.

    :meta hide-value:
    """

    reference_fee_units: Optional[int] = None
    """
    The cost, in fee units, of the reference transaction. This field is required.

    :meta hide-value:
    """

    reserve_base: Optional[int] = None
    """
    The base reserve, in drops. This field is required.

    :meta hide-value:
    """

    reserve_increment: Optional[int] = None
    """
    The incremental reserve, in drops. This field is required.

    :meta hide-value:
    """

    # Required AFTER the XRPFees Amendment

    base_fee_drops: Optional[str] = None
    """
    The charge, in drops of XRP, for the reference transaction, as hex. (This is the
    transaction cost before scaling for load.) This field is required.

    :meta hide-value:
    """

    reserve_base_drops: Optional[str] = None
    """
    The base reserve, in drops. This field is required.

    :meta hide-value:
    """

    reserve_increment_drops: Optional[str] = None
    """
    The incremental reserve, in drops. This field is required.

    :meta hide-value:
    """

    transaction_type: PseudoTransactionType = field(
        default=PseudoTransactionType.SET_FEE,
        init=False,
    )



"""
Model objects for specific `types of Transactions
<https://xrpl.org/transaction-types.html>`_ in the XRP Ledger.
"""

from xrpl.models.transactions.account_delete import AccountDelete
from xrpl.models.transactions.account_set import (
    AccountSet,
    AccountSetAsfFlag,
    AccountSetFlag,
    AccountSetFlagInterface,
)
from xrpl.models.transactions.amm_bid import AMMBid, AuthAccount
from xrpl.models.transactions.amm_clawback import AMMClawback
from xrpl.models.transactions.amm_create import AMMCreate
from xrpl.models.transactions.amm_delete import AMMDelete
from xrpl.models.transactions.amm_deposit import (
    AMMDeposit,
    AMMDepositFlag,
    AMMDepositFlagInterface,
)
from xrpl.models.transactions.amm_vote import AMMVote
from xrpl.models.transactions.amm_withdraw import (
    AMMWithdraw,
    AMMWithdrawFlag,
    AMMWithdrawFlagInterface,
)
from xrpl.models.transactions.batch import Batch, BatchFlag, BatchFlagInterface
from xrpl.models.transactions.check_cancel import CheckCancel
from xrpl.models.transactions.check_cash import CheckCash
from xrpl.models.transactions.check_create import CheckCreate
from xrpl.models.transactions.clawback import Clawback
from xrpl.models.transactions.credential_accept import CredentialAccept
from xrpl.models.transactions.credential_create import CredentialCreate
from xrpl.models.transactions.credential_delete import CredentialDelete
from xrpl.models.transactions.delegate_set import DelegateSet, GranularPermission
from xrpl.models.transactions.deposit_preauth import DepositPreauth
from xrpl.models.transactions.did_delete import DIDDelete
from xrpl.models.transactions.did_set import DIDSet
from xrpl.models.transactions.escrow_cancel import EscrowCancel
from xrpl.models.transactions.escrow_create import EscrowCreate
from xrpl.models.transactions.escrow_finish import EscrowFinish
from xrpl.models.transactions.metadata import TransactionMetadata
from xrpl.models.transactions.mptoken_authorize import (
    MPTokenAuthorize,
    MPTokenAuthorizeFlag,
    MPTokenAuthorizeFlagInterface,
)
from xrpl.models.transactions.mptoken_issuance_create import (
    MPTokenIssuanceCreate,
    MPTokenIssuanceCreateFlag,
    MPTokenIssuanceCreateFlagInterface,
)
from xrpl.models.transactions.mptoken_issuance_destroy import MPTokenIssuanceDestroy
from xrpl.models.transactions.mptoken_issuance_set import (
    MPTokenIssuanceSet,
    MPTokenIssuanceSetFlag,
    MPTokenIssuanceSetFlagInterface,
)
from xrpl.models.transactions.nftoken_accept_offer import NFTokenAcceptOffer
from xrpl.models.transactions.nftoken_burn import NFTokenBurn
from xrpl.models.transactions.nftoken_cancel_offer import NFTokenCancelOffer
from xrpl.models.transactions.nftoken_create_offer import (
    NFTokenCreateOffer,
    NFTokenCreateOfferFlag,
    NFTokenCreateOfferFlagInterface,
)
from xrpl.models.transactions.nftoken_mint import (
    NFTokenMint,
    NFTokenMintFlag,
    NFTokenMintFlagInterface,
)
from xrpl.models.transactions.nftoken_modify import NFTokenModify
from xrpl.models.transactions.offer_cancel import OfferCancel
from xrpl.models.transactions.offer_create import (
    OfferCreate,
    OfferCreateFlag,
    OfferCreateFlagInterface,
)
from xrpl.models.transactions.oracle_delete import OracleDelete
from xrpl.models.transactions.oracle_set import OracleSet
from xrpl.models.transactions.payment import Payment, PaymentFlag, PaymentFlagInterface
from xrpl.models.transactions.payment_channel_claim import (
    PaymentChannelClaim,
    PaymentChannelClaimFlag,
    PaymentChannelClaimFlagInterface,
)
from xrpl.models.transactions.payment_channel_create import PaymentChannelCreate
from xrpl.models.transactions.payment_channel_fund import PaymentChannelFund
from xrpl.models.transactions.permissioned_domain_delete import PermissionedDomainDelete
from xrpl.models.transactions.permissioned_domain_set import PermissionedDomainSet
from xrpl.models.transactions.set_regular_key import SetRegularKey
from xrpl.models.transactions.signer_list_set import SignerEntry, SignerListSet
from xrpl.models.transactions.ticket_create import TicketCreate
from xrpl.models.transactions.transaction import (
    Memo,
    Signer,
    Transaction,
    TransactionFlag,
    TransactionFlagInterface,
)
from xrpl.models.transactions.trust_set import (
    TrustSet,
    TrustSetFlag,
    TrustSetFlagInterface,
)
from xrpl.models.transactions.xchain_account_create_commit import (
    XChainAccountCreateCommit,
)
from xrpl.models.transactions.xchain_add_account_create_attestation import (
    XChainAddAccountCreateAttestation,
)
from xrpl.models.transactions.xchain_add_claim_attestation import (
    XChainAddClaimAttestation,
)
from xrpl.models.transactions.xchain_claim import XChainClaim
from xrpl.models.transactions.xchain_commit import XChainCommit
from xrpl.models.transactions.xchain_create_bridge import XChainCreateBridge
from xrpl.models.transactions.xchain_create_claim_id import XChainCreateClaimID
from xrpl.models.transactions.xchain_modify_bridge import (
    XChainModifyBridge,
    XChainModifyBridgeFlag,
    XChainModifyBridgeFlagInterface,
)

__all__ = [
    "AccountDelete",
    "AccountSet",
    "AccountSetAsfFlag",
    "AccountSetFlag",
    "AccountSetFlagInterface",
    "AMMBid",
    "AMMClawback",
    "AMMCreate",
    "AMMDelete",
    "AMMDeposit",
    "AMMDepositFlag",
    "AMMDepositFlagInterface",
    "AMMVote",
    "AMMWithdraw",
    "AMMWithdrawFlag",
    "AMMWithdrawFlagInterface",
    "AuthAccount",
    "Batch",
    "BatchFlag",
    "BatchFlagInterface",
    "CheckCancel",
    "CheckCash",
    "CheckCreate",
    "Clawback",
    "CredentialAccept",
    "CredentialCreate",
    "CredentialDelete",
    "DepositPreauth",
    "DelegateSet",
    "DIDDelete",
    "DIDSet",
    "EscrowCancel",
    "EscrowCreate",
    "EscrowFinish",
    "GranularPermission",
    "Memo",
    "MPTokenAuthorize",
    "MPTokenAuthorizeFlag",
    "MPTokenAuthorizeFlagInterface",
    "MPTokenIssuanceCreate",
    "MPTokenIssuanceCreateFlag",
    "MPTokenIssuanceCreateFlagInterface",
    "MPTokenIssuanceDestroy",
    "MPTokenIssuanceSet",
    "MPTokenIssuanceSetFlag",
    "MPTokenIssuanceSetFlagInterface",
    "NFTokenAcceptOffer",
    "NFTokenBurn",
    "NFTokenCancelOffer",
    "NFTokenCreateOffer",
    "NFTokenCreateOfferFlag",
    "NFTokenCreateOfferFlagInterface",
    "NFTokenMint",
    "NFTokenMintFlag",
    "NFTokenMintFlagInterface",
    "NFTokenModify",
    "OfferCancel",
    "OfferCreate",
    "OfferCreateFlag",
    "OfferCreateFlagInterface",
    "OracleDelete",
    "OracleSet",
    "Payment",
    "PaymentChannelClaim",
    "PaymentChannelClaimFlag",
    "PaymentChannelClaimFlagInterface",
    "PaymentChannelCreate",
    "PaymentChannelFund",
    "PaymentFlag",
    "PaymentFlagInterface",
    "PermissionedDomainDelete",
    "PermissionedDomainSet",
    "SetRegularKey",
    "Signer",
    "SignerEntry",
    "SignerListSet",
    "TicketCreate",
    "Transaction",
    "TransactionFlag",
    "TransactionFlagInterface",
    "TransactionMetadata",
    "TrustSet",
    "TrustSetFlag",
    "TrustSetFlagInterface",
    "XChainAccountCreateCommit",
    "XChainAddAccountCreateAttestation",
    "XChainAddClaimAttestation",
    "XChainClaim",
    "XChainCommit",
    "XChainCreateBridge",
    "XChainCreateClaimID",
    "XChainModifyBridge",
    "XChainModifyBridgeFlag",
    "XChainModifyBridgeFlagInterface",
]



"""Model for CheckCancel transaction type."""

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CheckCancel(Transaction):
    """
    Represents a `CheckCancel <https://xrpl.org/checkcancel.html>`_ transaction,
    which cancels an unredeemed Check, removing it from the ledger
    without sending any money. The source or the destination of the check
    can cancel a Check at any time using this transaction type. If the
    Check has expired, any address can cancel it.
    """

    check_id: str = REQUIRED  # type: ignore
    """
    The ID of the `Check ledger object
    <https://xrpl.org/check.html>`_ to cancel, as a 64-character
    hexadecimal string. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CHECK_CANCEL,
        init=False,
    )



"""
The sign_for command provides one signature for a multi-signed transaction.

By default, this method is admin-only. It can be used as a public method if the server
has enabled public signing.

This command requires the MultiSign amendment to be enabled.

`See sign_for <https://xrpl.org/sign_for.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type

from typing_extensions import Self

from xrpl.constants import CryptoAlgorithm
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class SignFor(Request):
    """
    The sign_for command provides one signature for a multi-signed transaction.

    By default, this method is admin-only. It can be used as a public method if the
    server has enabled public signing.

    This command requires the MultiSign amendment to be enabled.

    `See sign_for <https://xrpl.org/sign_for.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SIGN_FOR, init=False)
    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    transaction: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    secret: Optional[str] = None
    seed: Optional[str] = None
    seed_hex: Optional[str] = None
    passphrase: Optional[str] = None
    key_type: Optional[CryptoAlgorithm] = None

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new SignFor from a dictionary of parameters.

        Args:
            value: The value to construct the SignFor from.

        Returns:
            A new SignFor object, constructed using the given parameters.
        """
        if "tx_json" in value:
            fixed_value = {
                **value,
                "transaction": Transaction.from_xrpl(value["tx_json"]),
            }
            del fixed_value["tx_json"]
        else:
            fixed_value = value
        return super().from_dict(fixed_value)

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a SignFor.

        Returns:
            The dictionary representation of a SignFor.
        """
        return_dict = super().to_dict()
        del return_dict["transaction"]
        return_dict["tx_json"] = self.transaction.to_xrpl()
        return return_dict

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if not self._has_only_one_seed():
            errors["SignFor"] = (
                "Must have only one of `secret`, `seed`, `seed_hex`, and `passphrase`."
            )

        if self.secret is not None and self.key_type is not None:
            errors["key_type"] = "Must omit `key_type` if `secret` is provided."

        return errors

    def _has_only_one_seed(self: Self) -> bool:
        present_items = [
            item
            for item in [self.secret, self.seed, self.seed_hex, self.passphrase]
            if item is not None
        ]
        return len(present_items) == 1



"""
The random command provides a random number to be
used as a source of entropy for random number generation by clients.
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Random(Request):
    """
    The random command provides a random number to be
    used as a source of entropy for random number generation by clients.
    """

    method: RequestMethod = field(default=RequestMethod.RANDOM, init=False)



"""This method retrieves all of the NFTs currently owned by the specified account."""

from dataclasses import dataclass, field
from typing import Any, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AccountNFTs(Request, LookupByLedgerRequest):
    """
    This method retrieves all of the NFTs currently owned
    by the specified account.
    """

    method: RequestMethod = field(default=RequestMethod.ACCOUNT_NFTS, init=False)
    account: str = REQUIRED  # type: ignore
    """
    The unique identifier of an account, typically the account's address. The
    request returns NFTs owned by this account. This value is required.

    :meta hide-value:
    """

    limit: Optional[int] = None
    """Limit the number of NFTokens to retrieve."""

    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None
    """
    Value from a previous paginated response. Resume retrieving data where
    that response left off.
    """



"""Model for AMMWithdraw transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.amounts import Amount, IssuedCurrencyAmount
from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class AMMWithdrawFlag(int, Enum):
    """
    Transactions of the AMMWithdraw type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_LP_TOKEN = 0x00010000
    TF_WITHDRAW_ALL = 0x00020000
    TF_ONE_ASSET_WITHDRAW_ALL = 0x00040000
    TF_SINGLE_ASSET = 0x00080000
    TF_TWO_ASSET = 0x00100000
    TF_ONE_ASSET_LP_TOKEN = 0x00200000
    TF_LIMIT_LP_TOKEN = 0x00400000


class AMMWithdrawFlagInterface(TransactionFlagInterface):
    """
    Transactions of the AMMWithdraw type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_LP_TOKEN: bool
    TF_WITHDRAW_ALL: bool
    TF_ONE_ASSET_WITHDRAW_ALL: bool
    TF_SINGLE_ASSET: bool
    TF_TWO_ASSET: bool
    TF_ONE_ASSET_LP_TOKEN: bool
    TF_LIMIT_LP_TOKEN: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AMMWithdraw(Transaction):
    """
    Withdraw assets from an Automated Market Maker (AMM) instance by returning the
    AMM's liquidity provider tokens (LP Tokens).
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    The definition for one of the assets in the AMM's pool. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The definition for the other asset in the AMM's pool. This field is required.
    """

    amount: Optional[Amount] = None
    """
    The amount of one asset to withdraw from the AMM.
    This must match the type of one of the assets (tokens or XRP) in the AMM's pool.
    """

    amount2: Optional[Amount] = None
    """
    The amount of another asset to withdraw from the AMM.
    If present, this must match the type of the other asset in the AMM's pool
    and cannot be the same type as Amount.
    """

    e_price: Optional[Amount] = None
    """
    The minimum effective price, in LP Token returned, to pay per unit of the asset
    to withdraw.
    """

    lp_token_in: Optional[IssuedCurrencyAmount] = None
    """
    How many of the AMM's LP Tokens to redeem.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_WITHDRAW,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.amount2 is not None and self.amount is None:
            errors["AMMWithdraw"] = "Must set `amount` with `amount2`"
        elif self.e_price is not None and self.amount is None:
            errors["AMMWithdraw"] = "Must set `amount` with `e_price`"
        return errors



"""
This request returns the raw ledger format for all objects owned by an account.

For a higher-level view of an account's trust lines and balances, see
AccountLinesRequest instead.

`See account_objects <https://xrpl.org/account_objects.html>`_
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class AccountObjectType(str, Enum):
    """Represents the object types that an AccountObjectsRequest can ask for."""

    AMM = "amm"
    BRIDGE = "bridge"
    CHECK = "check"
    CREDENTIAL = "credential"
    DEPOSIT_PREAUTH = "deposit_preauth"
    DELEGATE = "delegate"
    DID = "did"
    ESCROW = "escrow"
    MPT_ISSUANCE = "mpt_issuance"
    MPTOKEN = "mptoken"
    NFT_OFFER = "nft_offer"
    OFFER = "offer"
    ORACLE = "oracle"
    PAYMENT_CHANNEL = "payment_channel"
    PERMISSIONED_DOMAIN = "permissioned_domain"
    SIGNER_LIST = "signer_list"
    STATE = "state"
    TICKET = "ticket"
    XCHAIN_OWNED_CREATE_ACCOUNT_CLAIM_ID = "xchain_owned_create_account_claim_id"
    XCHAIN_OWNED_CLAIM_ID = "xchain_owned_claim_id"


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AccountObjects(Request, LookupByLedgerRequest):
    """
    This request returns the raw ledger format for all objects owned by an account.

    For a higher-level view of an account's trust lines and balances, see
    AccountLinesRequest instead.

    `See account_objects <https://xrpl.org/account_objects.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.ACCOUNT_OBJECTS, init=False)
    type: Optional[AccountObjectType] = None
    deletion_blockers_only: bool = False
    limit: Optional[int] = None
    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None



"""Miscellaneous functions that are private to xrpl.core.keypairs."""

import hashlib

import Crypto.Hash.RIPEMD160 as RIPEMD160


def sha512_first_half(message: bytes) -> bytes:
    """
    Returns the first 32 bytes of SHA-512 hash of message.

    Args:
        message: Bytes input to hash.

    Returns:
        The first 32 bytes of SHA-512 hash of message.
    """
    return hashlib.sha512(message).digest()[:32]


def get_account_id(public_key: bytes) -> bytes:
    """
    Returns the account ID for a given public key. See
    https://xrpl.org/cryptographic-keys.html#account-id-and-address
    to learn about the relationship between keys and account IDs.

    Args:
        public_key: Unencoded public key.

    Returns:
        The account ID for the given public key.
    """
    sha_hash = hashlib.sha256(public_key).digest()
    return bytes(RIPEMD160.new(sha_hash).digest())



"""
The unsubscribe command tells the server to stop sending
messages for a particular subscription or set of subscriptions.

WebSocket API only.

`See unsubscribe <https://xrpl.org/unsubscribe.html>`_
"""

from dataclasses import dataclass, field
from typing import List, Optional

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies import Currency
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.requests.subscribe import StreamParameter
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class UnsubscribeBook(BaseModel):
    """Format for elements in the ``books`` array for Unsubscribe only."""

    taker_gets: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    taker_pays: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    both: bool = False


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Unsubscribe(Request):
    """
    The unsubscribe command tells the server to stop sending
    messages for a particular subscription or set of subscriptions.

    WebSocket API only.

    `See unsubscribe <https://xrpl.org/unsubscribe.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.UNSUBSCRIBE, init=False)
    streams: Optional[List[StreamParameter]] = None
    accounts: Optional[List[str]] = None
    accounts_proposed: Optional[List[str]] = None
    books: Optional[List[UnsubscribeBook]] = None



"""
Retrieve information about the public ledger.
`See ledger <https://xrpl.org/ledger.html>`_
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Ledger(Request, LookupByLedgerRequest):
    """
    Retrieve information about the public ledger.
    `See ledger <https://xrpl.org/ledger.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER, init=False)
    transactions: bool = False
    expand: bool = False
    owner_funds: bool = False
    binary: bool = False
    queue: bool = False



"""
Codec for serializing and deserializing Amount fields.
See `Amount Fields <https://xrpl.org/serialization.html#amount-fields>`_
"""

from __future__ import annotations

from decimal import MAX_PREC, Context, Decimal, InvalidOperation, localcontext
from typing import Any, Dict, Optional, Type, Union

from typing_extensions import Final, Self

from xrpl.constants import (
    IOU_DECIMAL_CONTEXT,
    MAX_IOU_EXPONENT,
    MAX_IOU_MANTISSA,
    MAX_IOU_PRECISION,
    MIN_IOU_EXPONENT,
    MIN_IOU_MANTISSA,
)
from xrpl.core.binarycodec.binary_wrappers import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.account_id import _HEX_REGEX, AccountID
from xrpl.core.binarycodec.types.currency import Currency
from xrpl.core.binarycodec.types.hash192 import Hash192
from xrpl.core.binarycodec.types.serialized_type import SerializedType
from xrpl.models.amounts import IssuedCurrencyAmount, MPTAmount

_MAX_DROPS: Final[Decimal] = Decimal("1e17")
_MIN_XRP: Final[Decimal] = Decimal("1e-6")

# other constants:
_NOT_XRP_BIT_MASK: Final[int] = 0x80
_POS_SIGN_BIT_MASK: Final[int] = 0x4000000000000000
_ZERO_CURRENCY_AMOUNT_HEX: Final[int] = 0x8000000000000000
_NATIVE_AMOUNT_BYTE_LENGTH: Final[int] = 8
_CURRENCY_AMOUNT_BYTE_LENGTH: Final[int] = 48
_MPT_MASK: Final[Decimal] = Decimal(0x8000000000000000)


def _contains_decimal(string: str) -> bool:
    """Returns True if the given string contains a decimal point character.

    Args:
        string: The string to check.

    Returns:
        True if the string contains a decimal point character.
    """
    return string.find(".") == -1


def verify_xrp_value(xrp_value: str) -> None:
    """
    Validates the format of an XRP amount.
    Raises if value is invalid.

    Args:
        xrp_value: A string representing an amount of XRP.

    Returns:
        None, but raises if xrp_value is not a valid XRP amount.

    Raises:
        XRPLBinaryCodecException: If xrp_value is not a valid XRP amount.
    """
    # Contains no decimal point
    if not _contains_decimal(xrp_value):
        raise XRPLBinaryCodecException(f"{xrp_value} is an invalid XRP amount.")

    # Within valid range
    decimal = Decimal(xrp_value)
    # Zero is less than both the min and max XRP amounts but is valid.
    if decimal.is_zero():
        return
    if (decimal.compare(_MIN_XRP) == -1) or (decimal.compare(_MAX_DROPS) == 1):
        raise XRPLBinaryCodecException(f"{xrp_value} is an invalid XRP amount.")


def verify_iou_value(issued_currency_value: str) -> None:
    """
    Validates the format of an issued currency amount value.
    Raises if value is invalid.

    Args:
        issued_currency_value: A string representing the "value"
                               field of an issued currency amount.

    Returns:
        None, but raises if issued_currency_value is not valid.

    Raises:
        XRPLBinaryCodecException: If issued_currency_value is invalid.
    """
    decimal_value = Decimal(issued_currency_value)
    if decimal_value.is_zero():
        return
    exponent = decimal_value.as_tuple().exponent
    if not isinstance(exponent, int):  # NaN, sNaN, Infinity
        raise XRPLBinaryCodecException(f"Expected exponent to be int, is {exponent}")
    if (
        (_calculate_precision(issued_currency_value) > MAX_IOU_PRECISION)
        or (exponent > MAX_IOU_EXPONENT)
        or (exponent < MIN_IOU_EXPONENT)
    ):
        raise XRPLBinaryCodecException(
            "Decimal precision out of range for issued currency value."
        )
    _verify_no_decimal(decimal_value)


def verify_mpt_value(mpt_value: str) -> None:
    """
    Validates the format of an MPT amount.
    Raises if value is invalid.

    Args:
        mpt_value: A string representing an amount of XRP.

    Returns:
        None, but raises if mpt_value is not a valid MPT amount.

    Raises:
        XRPLBinaryCodecException: If mpt_value is not a valid MPT amount.
    """
    # Contains no decimal point
    if not _contains_decimal(mpt_value):
        raise XRPLBinaryCodecException(
            f"{mpt_value} is a decimal value. Please use string type to represent it."
        )

    decimal = None
    try:
        # Check if the value is hexadecimal
        if mpt_value.startswith("0x") or _HEX_REGEX.fullmatch(mpt_value):
            decimal = Decimal(int(mpt_value, 16))
        else:
            # Check if mpt_value can be converted to a Decimal and within valid range
            decimal = Decimal(mpt_value)
    except (InvalidOperation, ValueError):
        raise XRPLBinaryCodecException(f"{mpt_value} is not a valid MPT amount.")

    # Zero is less than both the min and max MPT amounts but is valid.
    if decimal.is_zero():
        return

    # Perform the bitwise AND operation to check the MSB
    if int(decimal) & int(_MPT_MASK) != 0:
        raise XRPLBinaryCodecException(f"{mpt_value} is an illegal amount")


def _calculate_precision(value: str) -> int:
    """Calculate the precision of given value as a string."""
    decimal_value = Decimal(value, Context(prec=MAX_PREC))
    if decimal_value == decimal_value.to_integral():
        return len(
            decimal_value.quantize(Decimal(1), context=Context(prec=MAX_PREC))
            .as_tuple()
            .digits
        )
    return len(decimal_value.normalize(Context()).as_tuple().digits)


def _verify_no_decimal(decimal: Decimal) -> None:
    """
    Ensure that the value after being multiplied by the exponent
    does not contain a decimal.

    :param decimal: A Decimal object.
    """
    actual_exponent = decimal.as_tuple().exponent
    exponent = Decimal("1e" + str(-(int(actual_exponent) - 15)))
    if actual_exponent == 0:
        int_number_string = "".join([str(d) for d in decimal.as_tuple().digits])
    else:
        # str(Decimal) uses sci notation by default... get around w/ string format
        int_number_string = "{:f}".format(decimal * exponent)
    if not _contains_decimal(int_number_string):
        raise XRPLBinaryCodecException("Decimal place found in int_number_str")


def _serialize_issued_currency_value(value: str) -> bytes:
    """
    Serializes the value field of an issued currency amount to its bytes representation.

    :param value: The value to serialize, as a string.
    :return: A bytes object encoding the serialized value.
    """
    verify_iou_value(value)
    decimal_value = Decimal(value)
    if decimal_value.is_zero():
        return _ZERO_CURRENCY_AMOUNT_HEX.to_bytes(8, byteorder="big")

    # Convert components to integers ---------------------------------------
    sign, digits, exp = decimal_value.as_tuple()
    mantissa = int("".join([str(d) for d in digits]))
    if not isinstance(exp, int):  # NaN, sNaN, Infinity
        raise XRPLBinaryCodecException(f"Expected exp to be int, is {exp}")

    # Canonicalize to expected range ---------------------------------------
    while mantissa < MIN_IOU_MANTISSA and exp > MIN_IOU_EXPONENT:
        mantissa *= 10
        exp -= 1

    while mantissa > MAX_IOU_MANTISSA:
        if exp >= MAX_IOU_EXPONENT:
            raise XRPLBinaryCodecException(
                f"Amount overflow in issued currency value {str(value)}"
            )
        mantissa //= 10
        exp += 1

    if exp < MIN_IOU_EXPONENT or mantissa < MIN_IOU_MANTISSA:
        # Round to zero
        _ZERO_CURRENCY_AMOUNT_HEX.to_bytes(8, byteorder="big", signed=False)

    if exp > MAX_IOU_EXPONENT or mantissa > MAX_IOU_MANTISSA:
        raise XRPLBinaryCodecException(
            f"Amount overflow in issued currency value {str(value)}"
        )

    # Convert to bytes -----------------------------------------------------
    serial = _ZERO_CURRENCY_AMOUNT_HEX  # "Not XRP" bit set
    if sign == 0:
        serial |= _POS_SIGN_BIT_MASK  # "Is positive" bit set
    serial |= (exp + 97) << 54  # next 8 bits are exponents
    serial |= mantissa  # last 54 bits are mantissa

    return serial.to_bytes(8, byteorder="big", signed=False)


def _serialize_xrp_amount(value: str) -> bytes:
    """Serializes an XRP amount.

    Args:
        value: A string representing a quantity of XRP.

    Returns:
        The bytes representing the serialized XRP amount.
    """
    verify_xrp_value(value)
    # set the "is positive" bit (this is backwards from usual two's complement!)
    value_with_pos_bit = int(value) | _POS_SIGN_BIT_MASK
    return value_with_pos_bit.to_bytes(8, byteorder="big")


def _serialize_issued_currency_amount(value: Dict[str, str]) -> bytes:
    """Serializes an issued currency amount.

    Args:
        value: A dictionary representing an issued currency amount

    Returns:
         The bytes representing the serialized issued currency amount.
    """
    amount_string = value["value"]
    amount_bytes = _serialize_issued_currency_value(amount_string)
    currency_bytes = bytes(Currency.from_value(value["currency"]))
    issuer_bytes = bytes(AccountID.from_value(value["issuer"]))
    return amount_bytes + currency_bytes + issuer_bytes


def _serialize_mpt_amount(value: Dict[str, str]) -> bytes:
    """Serializes an MPT amount.

    Args:
        value: A dictionary representing a quantity of MPT.

    Returns:
        The bytes representing the serialized MPT amount.
    """
    amount_string = value["value"]
    verify_mpt_value(amount_string)

    # Convert the MPT amount string to a 64-bit integer and then to bytes
    decimal_value = None
    if amount_string.startswith("0x") or _HEX_REGEX.fullmatch(amount_string):
        decimal_value = Decimal(int(amount_string, 16))
    else:
        decimal_value = Decimal(amount_string)

    amount_bytes = int(decimal_value).to_bytes(8, byteorder="big", signed=False)

    # Create a bytearray for the mpt_issuance_id and serialize it
    mpt_issuance_id = bytearray()
    Hash192.from_value(value["mpt_issuance_id"]).to_byte_sink(mpt_issuance_id)

    # Create the leading byte and set the MPT leading byte
    leading_byte = bytearray(1)
    leading_byte[0] |= 0x60  # Set MPT leading byte

    # Concatenate the bytes
    return bytes(leading_byte + amount_bytes + mpt_issuance_id)


class Amount(SerializedType):
    """Codec for serializing and deserializing Amount fields.
    See `Amount Fields <https://xrpl.org/serialization.html#amount-fields>`_
    """

    def __init__(self: Self, buffer: bytes) -> None:
        """Construct an Amount from given bytes."""
        super().__init__(buffer)

    @classmethod
    def from_value(cls: Type[Self], value: Union[str, Dict[str, str]]) -> Self:
        """
        Construct an Amount from an issued currency amount or (for XRP),
        a string amount.

        See `Amount Fields <https://xrpl.org/serialization.html#amount-fields>`_

        Args:
            value: The value from which to construct an Amount.

        Returns:
            An Amount object.

        Raises:
            XRPLBinaryCodecException: if an Amount cannot be constructed.
        """
        with localcontext(IOU_DECIMAL_CONTEXT):
            if isinstance(value, str):
                return cls(_serialize_xrp_amount(value))
            if IssuedCurrencyAmount.is_dict_of_model(value):
                return cls(_serialize_issued_currency_amount(value))
            if MPTAmount.is_dict_of_model(value):
                return cls(_serialize_mpt_amount(value))

        raise XRPLBinaryCodecException(
            "Invalid type to construct an Amount: expected str or dict,"
            f" received {value.__class__.__name__}."
        )

    @classmethod
    def from_parser(
        cls: Type[Self], parser: BinaryParser, length_hint: Optional[int] = None
    ) -> Self:
        """Construct an Amount from an existing BinaryParser.

        Args:
            parser: The parser to construct the Amount object from.
            length_hint: Unused.

        Returns:
            An Amount object.
        """
        first_byte = parser.peek()

        is_iou = (first_byte & 0x80) != 0  # type: ignore
        if is_iou:
            return cls(parser.read(_CURRENCY_AMOUNT_BYTE_LENGTH))

        # the amount can be either MPT or XRP at this point
        is_mpt = (first_byte & 0x20) != 0  # type: ignore
        num_bytes = 33 if is_mpt else 8
        return cls(parser.read(num_bytes))

    def to_json(self: Self) -> Union[str, Dict[Any, Any]]:
        """Construct a JSON object representing this Amount.

        Returns:
            The JSON representation of this amount.

        Raises:
            XRPLBinaryCodecException: if invalid amount type for JSON conversion.
        """
        if self.is_native():
            sign = "" if self.is_positive() else "-"
            masked_bytes = (
                int.from_bytes(self.buffer, byteorder="big") & 0x3FFFFFFFFFFFFFFF
            )
            return f"{sign}{masked_bytes}"

        if self.is_iou():
            parser = BinaryParser(str(self))
            value_bytes = parser.read(8)
            currency = Currency.from_parser(parser)
            issuer = AccountID.from_parser(parser)
            b1 = value_bytes[0]
            b2 = value_bytes[1]
            is_positive = b1 & 0x40
            sign = "" if is_positive else "-"
            exponent = ((b1 & 0x3F) << 2) + ((b2 & 0xFF) >> 6) - 97
            hex_mantissa = hex(b2 & 0x3F) + value_bytes[2:].hex()
            int_mantissa = int(hex_mantissa[2:], 16)
            value = Decimal(f"{sign}{int_mantissa}") * Decimal(f"1e{exponent}")

            if value.is_zero():
                value_str = "0"
            else:
                value_str = str(value).rstrip("0").rstrip(".")
            verify_iou_value(value_str)

            return {
                "value": value_str,
                "currency": currency.to_json(),
                "issuer": issuer.to_json(),
            }

        if self.is_mpt():
            parser = BinaryParser(self.to_hex())
            leading_byte = parser.read(1)
            value_bytes = parser.read(8)
            mpt_issuance_id = Hash192.from_parser(parser)

            is_positive = leading_byte[0] & 0x40
            sign = "" if is_positive else "-"

            msb = int.from_bytes(value_bytes[:4], byteorder="big")
            lsb = int.from_bytes(value_bytes[4:], byteorder="big")
            num = (msb << 32) | lsb

            return {
                "value": f"{sign}{num}",
                "mpt_issuance_id": mpt_issuance_id.to_hex(),
            }

        raise XRPLBinaryCodecException("Invalid amount type for JSON conversion")

    def is_native(self: Self) -> bool:
        """Returns True if this amount is a native XRP amount.

        Returns:
            True if this amount is a native XRP amount, False otherwise.
        """
        # A native amount is one where both the IOU bit (0x80)
        # and MPT bit (0x20) are not set
        return (self.buffer[0] & 0x80) == 0 and (self.buffer[0] & 0x20) == 0

    def is_iou(self: Self) -> bool:
        """Returns True if this amount is an IOU amount.

        Returns:
            True if this amount is an IOU amount, False otherwise.
        """
        return (self.buffer[0] & 0x80) != 0

    def is_mpt(self: Self) -> bool:
        """Returns True if this amount is an MPT amount.

        Returns:
            True if this amount is an MPT amount, False otherwise.
        """
        # An MPT amount is one where the MPT bit (0x20) is set
        # and the IOU bit (0x80) is not set
        return (self.buffer[0] & 0x20) != 0 and (self.buffer[0] & 0x80) == 0

    def is_positive(self: Self) -> bool:
        """Returns True if 2nd bit in 1st byte is set to 1 (positive amount).

        Returns:
            True if 2nd bit in 1st byte is set to 1 (positive amount),
            False otherwise.
        """
        return (bytes(self)[0] & 0x40) > 0



"""
The `nft_buy_offers` method retrieves all of buy offers
for the specified NFToken.
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class NFTBuyOffers(Request, LookupByLedgerRequest):
    """
    The `nft_buy_offers` method retrieves all of buy offers
    for the specified NFToken.
    """

    method: RequestMethod = field(default=RequestMethod.NFT_BUY_OFFERS, init=False)
    nft_id: str = REQUIRED  # type: ignore
    """
    The unique identifier of an NFToken.
    The request returns buy offers for this NFToken. This value is required.

    :meta hide-value:
    """



"""Async methods for working with transactions on the XRP Ledger."""

from xrpl.asyncio.transaction.main import (
    _calculate_fee_per_transaction_type,
    autofill,
    autofill_and_sign,
    sign,
    sign_and_submit,
    simulate,
    submit,
    transaction_json_to_binary_codec_form,
)
from xrpl.asyncio.transaction.reliable_submission import (
    XRPLReliableSubmissionException,
    submit_and_wait,
)

__all__ = [
    "autofill",
    "autofill_and_sign",
    "sign",
    "sign_and_submit",
    "simulate",
    "submit",
    "submit_and_wait",
    "transaction_json_to_binary_codec_form",
    "XRPLReliableSubmissionException",
    "_calculate_fee_per_transaction_type",
]



"""High-level XRPL exports."""

from xrpl import account, clients, core, ledger, models, transaction, utils, wallet
from xrpl.constants import CryptoAlgorithm, XRPLException

__all__ = [
    "CryptoAlgorithm",
    "XRPLException",
    "account",
    "clients",
    "core",
    "ledger",
    "models",
    "transaction",
    "utils",
    "wallet",
]



"""Methods for working with transactions on the XRP Ledger."""

from xrpl.asyncio.transaction import (
    XRPLReliableSubmissionException,
    transaction_json_to_binary_codec_form,
)
from xrpl.transaction.batch_signers import (
    combine_batch_signers,
    sign_multiaccount_batch,
)
from xrpl.transaction.main import (
    autofill,
    autofill_and_sign,
    sign,
    sign_and_submit,
    simulate,
    submit,
)
from xrpl.transaction.multisign import multisign
from xrpl.transaction.reliable_submission import submit_and_wait

__all__ = [
    "autofill",
    "autofill_and_sign",
    "combine_batch_signers",
    "multisign",
    "sign",
    "sign_and_submit",
    "sign_multiaccount_batch",
    "simulate",
    "submit",
    "submit_and_wait",
    "transaction_json_to_binary_codec_form",
    "XRPLReliableSubmissionException",
]



"""
A path is an ordered array. Each member of a path is an
object that specifies the step.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.base_model import BaseModel
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class PathStep(BaseModel):
    """A PathStep represents an individual step along a Path."""

    account: Optional[str] = None
    currency: Optional[str] = None
    issuer: Optional[str] = None
    type: Optional[int] = None
    type_hex: Optional[str] = None

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "account": self._get_account_error(),
                "currency": self._get_currency_error(),
                "issuer": self._get_issuer_error(),
            }.items()
            if value is not None
        }

    def _get_account_error(self: Self) -> Optional[str]:
        if self.account is None:
            return None
        if self.currency is not None or self.issuer is not None:
            return "Cannot set account if currency or issuer are set"
        return None

    def _get_currency_error(self: Self) -> Optional[str]:
        if self.currency is None:
            return None
        if self.account is not None:
            return "Cannot set currency if account is set"
        if self.issuer is not None and self.currency.upper() == "XRP":
            return "Cannot set issuer if currency is XRP"
        return None

    def _get_issuer_error(self: Self) -> Optional[str]:
        if self.issuer is None:
            return None
        if self.account is not None:
            return "Cannot set issuer if account is set"
        if self.currency is not None and self.currency.upper() == "XRP":
            return "Cannot set issuer if currency is XRP"
        return None


Path = List[PathStep]
"""
A Path is an ordered array of PathSteps.
"""



"""Model for PermissionedDomainDelete transaction type."""

import re
from dataclasses import dataclass, field
from typing import Dict, Final, Pattern

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_DOMAIN_ID_REGEX: Final[Pattern[str]] = re.compile("[A-F0-9]{64}")
DOMAIN_ID_LENGTH: Final[int] = 64


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class PermissionedDomainDelete(Transaction):
    """This transaction deletes a PermissionedDomain object."""

    domain_id: str = REQUIRED  # type: ignore
    """The domain to delete."""

    transaction_type: TransactionType = field(
        default=TransactionType.PERMISSIONED_DOMAIN_DELETE,
        init=False,
    )
    """The transaction type (PermissionedDomainDelete)."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if len(self.domain_id) != DOMAIN_ID_LENGTH:
            errors["PermissionedDomainDelete"] = (
                f"domain_id must be {DOMAIN_ID_LENGTH} characters long."
            )
        elif not _DOMAIN_ID_REGEX.fullmatch(self.domain_id):
            errors["PermissionedDomainDelete"] = (
                "domain_id does not conform to hex format."
            )

        return errors



"""Model for UNLModify pseudo-transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.pseudo_transactions.pseudo_transaction import (
    PseudoTransaction,
)
from xrpl.models.transactions.types import PseudoTransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class UNLModify(PseudoTransaction):
    """
    A UNLModify pseudo-transaction marks a change to the `Negative UNL
    <https://xrpl.org/negative-unl.html>`_, indicating that a trusted validator has
    gone offline or come back online.
    """

    ledger_sequence: int = REQUIRED  # type: ignore
    """
    The ledger index where this pseudo-transaction appears. This distinguishes the
    pseudo-transaction from other occurrences of the same change.
    This field is required.

    :meta hide-value:
    """

    unl_modify_disabling: int = REQUIRED  # type: ignore
    """
    If 1, this change represents adding a validator to the Negative UNL. If 0, this
    change represents removing a validator from the Negative UNL. (No other values
    are allowed.) This field is required.

    :meta hide-value:
    """

    unl_modify_validator: str = REQUIRED  # type: ignore
    """
    The validator to add or remove, as identified by its master public key.
    This field is required.

    :meta hide-value:
    """

    transaction_type: PseudoTransactionType = field(
        default=PseudoTransactionType.UNL_MODIFY,
        init=False,
    )

    flags: Optional[Union[Dict[str, bool], int, List[int]]] = None
    """
    The Flags value of the EnableAmendment pseudo-transaction indicates the status
    of the amendment at the time of the ledger including the pseudo-transaction.
    A Flags value of 0 (no flags) or an omitted Flags field indicates that the
    amendment has been enabled, and applies to all ledgers afterward.
    """

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.unl_modify_disabling not in {0, 1}:
            errors["unl_modify_disabling"] = (
                "`unl_modify_disabling` is not equal to 0 or 1."
            )

        return errors



"""Model used in AMMBid transaction."""

from __future__ import annotations

from dataclasses import dataclass

from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AuthAccount(NestedModel):
    """Represents one entry in a list of AuthAccounts used in AMMBid transaction."""

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """



"""Model for AMMClawback transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.amounts import IssuedCurrencyAmount
from xrpl.models.currencies import Currency
from xrpl.models.currencies.issued_currency import IssuedCurrency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class AMMClawbackFlag(int, Enum):
    """
    Claw back the specified amount of Asset, and a corresponding amount of Asset2 based
    on the AMM pool's asset proportion; both assets must be issued by the issuer in the
    Account field. If this flag isn't enabled, the issuer claws back the specified
    amount of Asset, while a corresponding proportion of Asset2 goes back to the Holder.
    """

    TF_CLAW_TWO_ASSETS = 0x00000001


class AMMClawbackFlagInterface(TransactionFlagInterface):
    """
    Claw back the specified amount of Asset, and a corresponding amount of Asset2 based
    on the AMM pool's asset proportion; both assets must be issued by the issuer in the
    Account field. If this flag isn't enabled, the issuer claws back the specified
    amount of Asset, while a corresponding proportion of Asset2 goes back to the Holder.
    """

    TF_CLAW_TWO_ASSETS: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AMMClawback(Transaction):
    """
    Claw back tokens from a holder who has deposited your issued tokens into an AMM
    pool.
    """

    holder: str = REQUIRED  # type: ignore
    """The account holding the asset to be clawed back."""

    asset: IssuedCurrency = REQUIRED  # type: ignore
    """
    Specifies the asset that the issuer wants to claw back from the AMM pool. In JSON,
    this is an object with currency and issuer fields. The issuer field must match with
    Account.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other asset in the AMM's pool. In JSON, this is an object with
    currency and issuer fields (omit issuer for XRP).
    """

    amount: Optional[IssuedCurrencyAmount] = None
    """
    The maximum amount to claw back from the AMM account. The currency and issuer
    subfields should match the Asset subfields. If this field isn't specified, or the
    value subfield exceeds the holder's available tokens in the AMM, all of the
    holder's tokens are clawed back.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_CLAWBACK,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "AMMClawback": self._validate_wallet_and_amount_fields(),
            }.items()
            if value is not None
        }

    def _validate_wallet_and_amount_fields(self: Self) -> Optional[str]:
        errors = ""
        if self.account == self.holder:
            errors += "Issuer and holder wallets must be distinct."

        if self.account != self.asset.issuer:
            errors += (
                "Asset.issuer and AMMClawback transaction sender must be identical."
            )

        if self.amount is not None and (
            self.amount.issuer != self.asset.issuer
            or self.amount.currency != self.asset.currency
        ):
            errors += (
                "Amount.issuer and Amount.currency must match corresponding Asset "
                + "fields."
            )

        return errors if errors else None



"""Interface for all network clients to follow."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from typing_extensions import Final, Self

from xrpl.asyncio.clients.exceptions import XRPLRequestFailureException
from xrpl.models.requests import ServerInfo
from xrpl.models.requests.request import Request
from xrpl.models.response import Response

# The default request timeout duration. Set in Client._request_impl to allow more time
# for longer running commands.
REQUEST_TIMEOUT: Final[float] = 10.0


class Client(ABC):
    """
    Interface for all network clients to follow.

    :meta private:
    """

    def __init__(self: Self, url: str) -> None:
        """
        Initializes a client.

        Arguments:
            url: The url to which this client will connect
        """
        self.url = url
        self.network_id: Optional[int] = None
        self.build_version: Optional[str] = None

    @abstractmethod
    async def _request_impl(
        self: Self, request: Request, *, timeout: float = REQUEST_TIMEOUT
    ) -> Response:
        """
        This is the actual driver for a given Client's request. It must be
        async because all of the helper functions in this library are
        async-first. Implement this in a given Client.

        Arguments:
            request: An object representing information about a rippled request.
            timeout: The maximum tolerable delay on waiting for a response.

        Returns:
            The response from the server, as a Response object.

        :meta private:
        """
        pass


async def get_network_id_and_build_version(client: Client) -> None:
    """
    Get the network id and build version of the connected server.

    Args:
        client: The network client to use to send the request.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    # the required values are already present, no need for further processing
    if client.network_id and client.build_version:
        return

    response = await client._request_impl(ServerInfo())
    if response.is_successful():
        if "network_id" in response.result["info"]:
            client.network_id = response.result["info"]["network_id"]
        if not client.build_version and "build_version" in response.result["info"]:
            client.build_version = response.result["info"]["build_version"]
        return

    raise XRPLRequestFailureException(response.result)



"""Model for CheckCash transaction type."""

from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CheckCash(Transaction):
    """
    Represents a `CheckCash transaction <https://xrpl.org/checkcash.html>`_,
    which redeems a Check object to receive up to the amount authorized by the
    corresponding CheckCreate transaction. Only the Destination address of a
    Check can cash it.
    """

    check_id: str = REQUIRED  # type: ignore
    """
    The ID of the `Check ledger object
    <https://xrpl.org/check.html>`_ to cash, as a 64-character
    hexadecimal string. This field is required.

    :meta hide-value:
    """

    amount: Optional[Amount] = None
    """
    Redeem the Check for exactly this amount, if possible. The currency must
    match that of the SendMax of the corresponding CheckCreate transaction.
    You must provide either this field or ``DeliverMin``.
    """

    deliver_min: Optional[Amount] = None
    """
    Redeem the Check for at least this amount and for as much as possible.
    The currency must match that of the ``SendMax`` of the corresponding
    CheckCreate transaction. You must provide either this field or ``Amount``.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CHECK_CASH,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if not (self.amount is None) ^ (self.deliver_min is None):
            errors["CheckCash"] = (
                "either amount or deliver_min must be set but not both"
            )
        return errors



"""Model for a XChainCreateClaimID transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from typing_extensions import Self

from xrpl.core.addresscodec import is_valid_classic_address
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainCreateClaimID(Transaction):
    """
    Represents a XChainCreateClaimID transaction.
    The XChainCreateClaimID transaction creates a new cross-chain claim ID that
    is used for a cross-chain transfer. A cross-chain claim ID represents one
    cross-chain transfer of value.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to create the claim ID for. This field is required.

    :meta hide-value:
    """

    signature_reward: str = REQUIRED  # type: ignore
    """
    The amount, in XRP, to reward the witness servers for providing signatures.
    This must match the amount on the ``Bridge`` ledger object. This field is
    required.

    :meta hide-value:
    """

    other_chain_source: str = REQUIRED  # type: ignore
    """
    The account that must send the corresponding ``XChainCommit`` transaction
    on the source chain. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_CREATE_CLAIM_ID,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.signature_reward != REQUIRED and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "`signature_reward` must be numeric."

        if self.other_chain_source != REQUIRED and not is_valid_classic_address(
            self.other_chain_source
        ):
            errors["other_chain_source"] = (
                "`other_chain_source` must be a valid XRPL address."
            )

        return errors



"""
Abstract base class for cryptographic algorithms in the XRP Ledger. The classes
for all cryptographic algorithms are derived from this interface.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Tuple, Type

from ecpy.keys import ECPrivateKey  # type: ignore
from typing_extensions import Self


class CryptoImplementation(ABC):
    """
    Abstract base class for cryptographic algorithms in the XRP Ledger. The
    classes for all cryptographic algorithms are derived from this interface.
    """

    @classmethod
    @abstractmethod
    def derive_keypair(  # noqa: D102
        cls: Type[Self],
        decoded_seed: bytes,
        is_validator: bool,
    ) -> Tuple[str, str]:
        pass

    @classmethod
    @abstractmethod
    def sign(  # noqa: D102
        cls: Type[Self],
        message: bytes,
        private_key: str,
    ) -> bytes:
        pass

    @classmethod
    @abstractmethod
    def is_valid_message(  # noqa: D102
        cls: Type[Self],
        message: bytes,
        signature: bytes,
        public_key: str,
    ) -> bool:
        pass

    @classmethod
    def _private_key_to_str(cls: Type[Self], key: ECPrivateKey) -> str:
        return format(key.d, "x")



"""General XRPL Client Exceptions."""

from __future__ import annotations

from typing import Any, Dict

from typing_extensions import Self

from xrpl.constants import XRPLException


class XRPLRequestFailureException(XRPLException):
    """XRPL Request Exception, when the request fails."""

    def __init__(self: Self, result: Dict[str, Any]) -> None:
        """
        Initializes a XRPLRequestFailureException.

        Args:
            result: the error result returned by the ledger.
        """
        self.error = result["error"]
        self.error_message = None
        if "error_message" in result and result["error_message"] is not None:
            self.error_message = result["error_message"]
        elif "error_exception" in result:
            self.error_message = result["error_exception"]
        self.message = f"Request failed, {self.error}: {self.error_message}"
        super().__init__(self.message)


class XRPLWebsocketException(XRPLException):
    """
    XRPL Websocket Exception. Thrown when something goes wrong with the Web Socket
    client.
    """

    pass



"""All transaction flags and utils to build a list of ints from a FlagInterface"""

from enum import Enum
from typing import Dict, List, Type, Union

from typing_extensions import TypedDict

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.transactions.types.pseudo_transaction_type import PseudoTransactionType
from xrpl.models.transactions.types.transaction_type import TransactionType


class FlagInterface(TypedDict):
    """A TypedDict to define transaction flags by bool."""

    pass


def _flag_enum_to_dict(flag_enum: Type[Enum]) -> Dict[str, int]:
    return {flag.name.lower(): flag.value for flag in flag_enum}


def interface_to_flag_list(
    tx_type: Union[TransactionType, PseudoTransactionType],
    tx_flags: Dict[str, bool],
) -> List[int]:
    """Parse a list of flags expressed as integers from the FlagInterface.

    Args:
        tx_type: Type of the transaction.
        tx_flags: FlagInterface

    Returns:
        List[int]:
            A list of flags expressed as integers.
    """
    from xrpl.models import transactions  # Avoid circular dependencies

    # Get all exported classes with names of the form `{TxType}Flag`
    # These are all the flag value enums for transactions/pseudo-transactions
    tx_flag_enums = [
        f for f in transactions.__all__ if f.endswith("Flag") and "Asf" not in f
    ]
    pseudo_tx_flag_enums = [
        f for f in transactions.pseudo_transactions.__all__ if f.endswith("Flag")
    ]

    # Key is transaction type name, value is mapping of flag name to int value
    all_tx_flags: Dict[str, Dict[str, int]] = {
        # The `:-4` here removes the `Flag` at the end of the class type to just get
        # the transaction type name
        **{f[:-4]: _flag_enum_to_dict(getattr(transactions, f)) for f in tx_flag_enums},
        **{
            f[:-4]: _flag_enum_to_dict(getattr(transactions.pseudo_transactions, f))
            for f in pseudo_tx_flag_enums
        },
    }

    # if transaction types has no flags
    if tx_type not in all_tx_flags:
        return [0]

    tx_specific_flags = all_tx_flags[tx_type]  # get flags for that transaction type
    flag_list = []  # accumulator of the flag numbers
    for flag, flag_on in tx_flags.items():  # iterate through user-included flags
        flag = flag.lower()  # normalize capitalization
        if flag_on:  # if flag was set to True
            if flag in tx_specific_flags:  # if flag was defined
                flag_list.append(tx_specific_flags[flag])
            else:
                flag_list.append(0)  # if flag was set to False
    return flag_list


def check_false_flag_definition(
    tx_type: Union[TransactionType, PseudoTransactionType],
    tx_flags: Union[Dict[str, bool], List[int]],
) -> None:
    """Check the flags were set correctly if not defined as integer.

    Args:
        tx_type (Union[TransactionType, PseudoTransactionType]):
            Type of the transaction.
        tx_flags (Iterable):
            FlagInterface

    Raises:
        XRPLModelException: Flags were not set correctly.
    """
    try:
        if isinstance(tx_flags, dict):
            interface_flags: Dict[str, bool] = tx_flags
            assert all(
                [
                    all(
                        (
                            isinstance(flag, str),
                            isinstance(set_flag, bool),
                        )
                    )
                    for flag, set_flag in interface_flags.items()
                ]
            )
        else:  # List[int]
            list_flags: List[int] = tx_flags
            assert all(isinstance(flag, int) for flag in list_flags)
    except AssertionError:
        msg = f"""
False flag definition: Please define flags either by setting bools using
`{tx_type}FlagInterface` or by using `{tx_type}Flag`.
Do not put the FlagInterface in a list or mix them.
""".strip()
        raise XRPLModelException(msg)



"""Maps and helpers providing serialization-related information about fields."""

import json
import os
from typing import Any, Dict, cast

from xrpl.core.binarycodec.definitions.field_header import FieldHeader
from xrpl.core.binarycodec.definitions.field_info import FieldInfo
from xrpl.core.binarycodec.definitions.field_instance import FieldInstance
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException


def load_definitions(filename: str = "definitions.json") -> Dict[str, Any]:
    """
    Loads JSON from the definitions file and converts it to a preferred format.
    The definitions file contains information required for the XRP Ledger's
    canonical binary serialization format:
    `Serialization <https://xrpl.org/serialization.html>`_

    Args:
        filename: The name of the definitions file.
            (The definitions file should be drop-in compatible with the one from the
            ripple-binary-codec JavaScript package.)

    Returns:
        A dictionary containing the mappings provided in the definitions file.
    """
    dirname = os.path.dirname(__file__)
    absolute_path = os.path.join(dirname, filename)
    with open(absolute_path) as definitions_file:
        definitions = json.load(definitions_file)
        return {
            "TYPES": definitions["TYPES"],
            # type_name str: type_sort_key int
            "FIELDS": {
                k: v for (k, v) in definitions["FIELDS"]
            },  # convert list of tuples to dict
            # "field_name" str: {
            #   "nth": field_sort_key int,
            #   "isVLEncoded": bool,
            #   "isSerialized": bool,
            #   "isSigningField": bool,
            #   "type": string
            # }
            "LEDGER_ENTRY_TYPES": definitions["LEDGER_ENTRY_TYPES"],
            "TRANSACTION_RESULTS": definitions["TRANSACTION_RESULTS"],
            "TRANSACTION_TYPES": definitions["TRANSACTION_TYPES"],
        }


_DEFINITIONS = load_definitions()
_TRANSACTION_TYPE_CODE_TO_STR_MAP = {
    value: key for (key, value) in _DEFINITIONS["TRANSACTION_TYPES"].items()
}
_TRANSACTION_RESULTS_CODE_TO_STR_MAP = {
    value: key for (key, value) in _DEFINITIONS["TRANSACTION_RESULTS"].items()
}
_LEDGER_ENTRY_TYPES_CODE_TO_STR_MAP = {
    value: key for (key, value) in _DEFINITIONS["LEDGER_ENTRY_TYPES"].items()
}

_GRANULAR_PERMISSIONS = {
    "TrustlineAuthorize": 65537,
    "TrustlineFreeze": 65538,
    "TrustlineUnfreeze": 65539,
    "AccountDomainSet": 65540,
    "AccountEmailHashSet": 65541,
    "AccountMessageKeySet": 65542,
    "AccountTransferRateSet": 65543,
    "AccountTickSizeSet": 65544,
    "PaymentMint": 65545,
    "PaymentBurn": 65546,
    "MPTokenIssuanceLock": 65547,
    "MPTokenIssuanceUnlock": 65548,
}

_tx_delegations = {
    key: value + 1 for (key, value) in _DEFINITIONS["TRANSACTION_TYPES"].items()
}
_DELEGABLE_PERMISSIONS_STR_TO_CODE_MAP: Dict[str, int] = {
    **_tx_delegations,
    **_GRANULAR_PERMISSIONS,
}
_DELEGABLE_PERMISSIONS_CODE_TO_STR_MAP: Dict[int, str] = {
    **{value: key for (key, value) in _DELEGABLE_PERMISSIONS_STR_TO_CODE_MAP.items()},
}

_TYPE_ORDINAL_MAP = _DEFINITIONS["TYPES"]

_FIELD_INFO_MAP = {}
_FIELD_HEADER_NAME_MAP: Dict[FieldHeader, str] = {}

# Populate _FIELD_INFO_MAP and _FIELD_HEADER_NAME_MAP
try:
    for field in _DEFINITIONS["FIELDS"]:
        field_entry = _DEFINITIONS["FIELDS"][field]
        field_info = FieldInfo(
            field_entry["nth"],
            field_entry["isVLEncoded"],
            field_entry["isSerialized"],
            field_entry["isSigningField"],
            field_entry["type"],
        )
        header = FieldHeader(_TYPE_ORDINAL_MAP[field_entry["type"]], field_entry["nth"])
        _FIELD_INFO_MAP[field] = field_info
        _FIELD_HEADER_NAME_MAP[header] = field
except KeyError as e:
    raise XRPLBinaryCodecException(
        f"Malformed definitions.json file. (Original exception: KeyError: {e})"
    )


def get_field_type_name(field_name: str) -> str:
    """
    Returns the serialization data type for the given field name.
    `Serialization Type List <https://xrpl.org/serialization.html#type-list>`_

    Args:
        field_name: The name of the field to get the serialization data type for.

    Returns:
        The serialization data type for the given field name.
    """
    return _FIELD_INFO_MAP[field_name].type


def get_field_type_code(field_name: str) -> int:
    """
    Returns the type code associated with the given field.
    `Serialization Type Codes <https://xrpl.org/serialization.html#type-codes>`_

    Args:
        field_name: The name of the field get a type code for.

    Returns:
        The type code associated with the given field name.

    Raises:
        XRPLBinaryCodecException: If definitions.json is invalid.
    """
    field_type_name = get_field_type_name(field_name)
    field_type_code = _TYPE_ORDINAL_MAP[field_type_name]
    if not isinstance(field_type_code, int):
        raise XRPLBinaryCodecException(
            "Field type codes in definitions.json must be ints."
        )

    return field_type_code


def get_field_code(field_name: str) -> int:
    """
    Returns the field code associated with the given field.
    `Serialization Field Codes <https://xrpl.org/serialization.html#field-codes>`_

    Args:
        field_name: The name of the field to get a field code for.

    Returns:
        The field code associated with the given field.
    """
    return _FIELD_INFO_MAP[field_name].nth


def get_field_header_from_name(field_name: str) -> FieldHeader:
    """
    Returns a FieldHeader object for a field of the given field name.

    Args:
        field_name: The name of the field to get a FieldHeader for.

    Returns:
        A FieldHeader object for a field of the given field name.
    """
    return FieldHeader(get_field_type_code(field_name), get_field_code(field_name))


def get_field_name_from_header(field_header: FieldHeader) -> str:
    """
    Returns the field name described by the given FieldHeader object.

    Args:
        field_header: The header to get a field name for.

    Returns:
        The name of the field described by the given FieldHeader.
    """
    return _FIELD_HEADER_NAME_MAP[field_header]


def get_field_instance(field_name: str) -> FieldInstance:
    """
    Return a FieldInstance object for the given field name.

    Args:
        field_name: The name of the field to get a FieldInstance for.

    Returns:
        A FieldInstance object for the given field name.
    """
    info = _FIELD_INFO_MAP[field_name]
    field_header = get_field_header_from_name(field_name)
    return FieldInstance(
        info,
        field_name,
        field_header,
    )


def get_transaction_type_code(transaction_type: str) -> int:
    """
    Return an integer representing the given transaction type string in an enum.

    Args:
        transaction_type: The name of the transaction type to get the enum value for.

    Returns:
        An integer representing the given transaction type string in an enum.
    """
    return cast(int, _DEFINITIONS["TRANSACTION_TYPES"][transaction_type])


def get_transaction_type_name(transaction_type: int) -> str:
    """
    Return string representing the given transaction type from the enum.

    Args:
        transaction_type: The enum value of the transaction type.

    Returns:
        The string name of the transaction type.
    """
    return cast(str, _TRANSACTION_TYPE_CODE_TO_STR_MAP[transaction_type])


def get_transaction_result_code(transaction_result_type: str) -> int:
    """
    Return an integer representing the given transaction result string in an enum.

    Args:
        transaction_result_type: The name of the transaction result type to get the
            enum value for.

    Returns:
        An integer representing the given transaction result type string in an enum.
    """
    return cast(int, _DEFINITIONS["TRANSACTION_RESULTS"][transaction_result_type])


def get_transaction_result_name(transaction_result_type: int) -> str:
    """
    Return string representing the given transaction result type from the enum.

    Args:
        transaction_result_type: The enum value of the transaction result type.

    Returns:
        The string name of the transaction result type.
    """
    return cast(str, _TRANSACTION_RESULTS_CODE_TO_STR_MAP[transaction_result_type])


def get_ledger_entry_type_code(ledger_entry_type: str) -> int:
    """
    Return an integer representing the given ledger entry type string in an enum.

    Args:
        ledger_entry_type: The name of the ledger entry type to get the enum value for.

    Returns:
        An integer representing the given ledger entry type string in an enum.
    """
    return cast(int, _DEFINITIONS["LEDGER_ENTRY_TYPES"][ledger_entry_type])


def get_ledger_entry_type_name(ledger_entry_type: int) -> str:
    """
    Return string representing the given ledger entry type from the enum.

    Args:
        ledger_entry_type: The enum value of the ledger entry type.

    Returns:
        The string name of the ledger entry type.
    """
    return cast(str, _LEDGER_ENTRY_TYPES_CODE_TO_STR_MAP[ledger_entry_type])


def get_permission_value_type_code(permission_value: str) -> int:
    """
    Return an integer representing the given permission value string.

    Args:
        permission_value: The name of the permission value to get the integer value for.

    Returns:
        An integer representing the given permission value string.
    """
    return _DELEGABLE_PERMISSIONS_STR_TO_CODE_MAP[permission_value]


def get_permission_value_type_name(permission_value: int) -> str:
    """
    Return string representing the given permission value from the integer.

    Args:
        permission_value: The integer permission value.

    Returns:
        The string name of the permission value.
    """
    return _DELEGABLE_PERMISSIONS_CODE_TO_STR_MAP[permission_value]



"""
The `nfts_by_issuer` method retrieves all of the NFTokens
issued by an account
"""

from dataclasses import dataclass, field
from typing import Any, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class NFTsByIssuer(Request, LookupByLedgerRequest):
    """
    The `nfts_by_issuer` method retrieves all of the NFTokens
    issued by an account
    """

    method: RequestMethod = field(default=RequestMethod.NFTS_BY_ISSUER, init=False)
    issuer: str = REQUIRED  # type: ignore
    """
    The unique identifier for an account that issues NFTokens
    The request returns NFTokens issued by this account. This field is required

    :meta hide-value:
    """

    marker: Optional[Any] = None
    nft_taxon: Optional[int] = None
    limit: Optional[int] = None



"""Model for Clawback transaction type and related flags."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional, Union

from typing_extensions import Self

from xrpl.models.amounts import (
    IssuedCurrencyAmount,
    MPTAmount,
    is_issued_currency,
    is_xrp,
)
from xrpl.models.amounts.amount import is_mpt
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Clawback(Transaction):
    """The clawback transaction claws back issued funds from token holders."""

    amount: Union[IssuedCurrencyAmount, MPTAmount] = REQUIRED  # type: ignore
    """
    The amount of currency to claw back. The issuer field is used for the token holder's
    address, from whom the tokens will be clawed back.

    :meta hide-value:
    """

    holder: Optional[str] = None
    """
    Indicates the AccountID that the issuer wants to clawback. This field is only valid
    for clawing back MPTs.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CLAWBACK,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        # Amount transaction errors
        if is_xrp(self.amount):
            errors["amount"] = "``amount`` cannot be XRP."
        elif is_issued_currency(self.amount):
            if self.holder is not None:
                errors["amount"] = "Cannot have Holder for currency."
            if self.account == self.amount.issuer:
                errors["amount"] = "Holder's address is wrong."
        elif is_mpt(self.amount):
            if self.holder is None:
                errors["amount"] = "Missing Holder."
            elif self.account == self.holder:
                errors["amount"] = "Invalid Holder account."

        return errors



"""
The XRP Ledger has three kinds of money: XRP, issued currencies, and MPTs. All types
have high precision, although their formats are different.
"""

from xrpl.models.currencies.currency import Currency
from xrpl.models.currencies.issued_currency import IssuedCurrency
from xrpl.models.currencies.mpt_currency import MPTCurrency
from xrpl.models.currencies.xrp import XRP

__all__ = [
    "Currency",
    "IssuedCurrency",
    "MPTCurrency",
    "XRP",
]



"""Utils to parse NFTokenIDs."""

from typing_extensions import TypedDict

from xrpl.constants import XRPLException
from xrpl.core.addresscodec.codec import encode_classic_address


class NFTokenID(TypedDict):
    """A decoded representation of info from the NFTokenID."""

    nftoken_id: str
    flags: int
    transfer_fee: int
    issuer: str
    taxon: int
    sequence: int


def unscramble_taxon(taxon: int, token_seq: int) -> int:
    """
    Unscrambles or rescrambles a taxon in an NFTokenID.

    An issuer may issue several NFTs with the same taxon; to ensure that NFTs
    are spread across multiple pages we lightly mix the taxon up by using the
    sequence (which is not under the issuer's direct control) as the seed for
    a simple linear congruential generator.

    From the Hull-Dobell theorem we know that f(x)=(m*x+c) mod n will yield a
    permutation of [0, n) when n is a power of 2 if m is congruent to 1 mod 4
    and c is odd. By doing a bitwise XOR with this permutation we can
    scramble/unscramble the taxon.

    The XLS-20d proposal fixes m = 384160001 and c = 2459.
    We then take the modulus of 2^32 which is 4294967296.

    Args:
        taxon: The scrambled or unscrambled taxon (The XOR is both the
                encoding and decoding)
        token_seq: The account sequence when the token was minted. Used as a
                pseudorandom seed.

    Returns:
        The opposite taxon. If the taxon was scrambled it becomes unscrambled,
        and vice versa.
    """
    return (taxon ^ (384160001 * token_seq + 2459)) % 4294967296


def parse_nftoken_id(nft_id: str) -> NFTokenID:
    """
    Parse an NFTokenID into the information it is encoding.

    Example decoding:

    000B 0539 C35B55AA096BA6D87A6E6C965A6534150DC56E5E 12C5D09E 0000000C
    +--- +--- +--------------------------------------- +------- +-------
    |    |    |                                        |        |
    |    |    |                                        |        `---> Sequence: 12
    |    |    |                                        |
    |    |    |                                        `---> Scrambled Taxon: 314,953,886
    |    |    |                                              Unscrambled Taxon: 1337
    |    |    |
    |    |    `---> Issuer: rJoxBSzpXhPtAuqFmqxQtGKjA13jUJWthE
    |    |
    |    `---> TransferFee: 1337.0 bps or 13.37%
    |
    `---> Flags: 11 -> lsfBurnable, lsfOnlyXRP and lsfTransferable

    Args:
        nft_id: A hex string which identifies an NFToken on the ledger.

    Raises:
        XRPLException: when given an invalid Token ID as nft_id.

    Returns:
        A decoded nft TokenID with all information encoded within

    # noqa:E501
    """
    expected_length = 64
    if len(nft_id) != expected_length:
        raise XRPLException(
            f"Attempting to parse a tokenID with length"
            f" {len(nft_id)}, but expected a length of {expected_length}`"
        )

    scrambled_taxon = int(nft_id[48:56], base=16)
    sequence = int(nft_id[56:64], base=16)

    nftoken_data: NFTokenID = {
        "nftoken_id": nft_id,
        "flags": int(nft_id[0:4], base=16),
        "transfer_fee": int(nft_id[4:8], base=16),
        "issuer": encode_classic_address(bytes.fromhex(nft_id[8:48])),
        "taxon": unscramble_taxon(scrambled_taxon, sequence),
        "sequence": sequence,
    }

    return nftoken_data



"""A sync client for interacting with the rippled JSON RPC."""

from xrpl.asyncio.clients.json_rpc_base import JsonRpcBase
from xrpl.clients.sync_client import SyncClient


class JsonRpcClient(SyncClient, JsonRpcBase):
    """A sync client for interacting with the rippled JSON RPC."""

    pass



"""Codec for serializing and deserializing a hash field with a width
of 192 bits (24 bytes).
`See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
"""

from __future__ import annotations

from typing import Type

from typing_extensions import Self

from xrpl.core.binarycodec.types.hash import Hash

HASH192_BYTES = 24


class Hash192(Hash):
    """
    Codec for serializing and deserializing a hash field with a width
    of 192 bits (24 bytes).
    `See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
    """

    @classmethod
    def _get_length(cls: Type[Self]) -> int:
        return 24



"""
The base class for all network response types.

Represents fields common to all response types.
"""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union

from typing_extensions import Self

from xrpl.models.base_model import BaseModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions import PaymentFlag
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class ResponseStatus(str, Enum):
    """Represents the different status possibilities."""

    SUCCESS = "success"
    ERROR = "error"


class ResponseType(str, Enum):
    """Represents the different response types a Response can have."""

    RESPONSE = "response"
    LEDGER_CLOSED = "ledgerClosed"
    TRANSACTION = "transaction"


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Response(BaseModel):
    """
    The base class for all network response types.

    Represents fields common to all response types.
    """

    status: ResponseStatus = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    result: Dict[str, Any] = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    id: Optional[Union[int, str]] = None
    type: Optional[ResponseType] = None

    def __post_init__(self: Self) -> None:
        """Called by dataclasses immediately after __init__."""
        super().__post_init__()
        if self.contains_partial_payment():
            warnings.warn(
                """This response contains a partial payment. Please confirm
                the delivered amount is correct""",
                stacklevel=2,
            )

    def is_successful(self: Self) -> bool:
        """
        Returns whether the request was successfully received and understood by the
        server.

        Returns:
            Whether the request was successfully received and understood by the server.
        """
        return self.status == ResponseStatus.SUCCESS

    def contains_partial_payment(self: Self) -> bool:
        """
        Returns whether the request contains at least one transactions with
        the partial payment flag set.

        Returns:
            True if at least one transaction in this Response has the partial
            payment flag set. False otherwise.
        """
        return self._do_contains_partial_payment(self.result)

    def _do_contains_partial_payment(self: Self, val: Any) -> bool:  # noqa: ANN401
        flagged = []
        if isinstance(val, dict):
            formatted = {key.strip().lower(): value for key, value in val.items()}
            if (
                "transactiontype" in formatted
                and formatted["transactiontype"] == TransactionType.PAYMENT
            ):
                flagged = [
                    True
                    for key, value in val.items()
                    if self._is_partial_payment(key, value)
                ]
        if isinstance(val, list):
            flagged = [
                True for sub_val in val if self._do_contains_partial_payment(sub_val)
            ]
        return len(flagged) > 0

    def _is_partial_payment(self: Self, key: str, val: Any) -> bool:  # noqa: ANN401
        if isinstance(val, dict):
            return self._do_contains_partial_payment(val)
        try:
            int_val = int(val)
        except (TypeError, ValueError):
            return False
        return key.strip().lower() == "flags" and (
            int_val & PaymentFlag.TF_PARTIAL_PAYMENT != 0
        )



"""Model for EscrowCreate transaction type."""

from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class EscrowCreate(Transaction):
    """
    Represents an `EscrowCreate <https://xrpl.org/escrowcreate.html>`_
    transaction, which locks up XRP until a specific time or condition is met.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    Amount of XRP, in drops, to deduct from the sender's balance and set
    aside in escrow. This field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The address that should receive the escrowed XRP when the time or
    condition is met. This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Escrow, or a hosted recipient to pay.
    """

    cancel_after: Optional[int] = None
    """
    The time, in seconds since the Ripple Epoch, when this escrow expires.
    This value is immutable; the funds can only be returned the sender after
    this time.
    """

    finish_after: Optional[int] = None
    """
    The time, in seconds since the Ripple Epoch, when the escrowed XRP can
    be released to the recipient. This value is immutable; the funds cannot
    move until this time is reached.
    """

    condition: Optional[str] = None
    """
    Hex value representing a `PREIMAGE-SHA-256 crypto-condition
    <https://tools.ietf.org/html/draft-thomas-crypto-conditions-04#section-8.1.>`_
    The funds can only be delivered to the recipient if this condition is
    fulfilled.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ESCROW_CREATE,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if (
            self.cancel_after is not None
            and self.finish_after is not None
            and self.finish_after >= self.cancel_after
        ):
            errors["EscrowCreate"] = (
                "The finish_after time must be before the cancel_after time."
            )

        return errors



"""
The submit_multisigned command applies a multi-signed transaction and sends it to the
network to be included in future ledgers. (You can also submit multi-signed
transactions in binary form using the submit command in submit-only mode.)

This command requires the MultiSign amendment to be enabled.

`See submit_multisigned <https://xrpl.org/submit_multisigned.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type

from typing_extensions import Self

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class SubmitMultisigned(Request):
    """
    The submit_multisigned command applies a multi-signed transaction and sends it to
    the network to be included in future ledgers. (You can also submit multi-signed
    transactions in binary form using the submit command in submit-only mode.)

    This command requires the MultiSign amendment to be enabled.

    `See submit_multisigned <https://xrpl.org/submit_multisigned.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SUBMIT_MULTISIGNED, init=False)
    tx_json: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    fail_hard: bool = False

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new SubmitMultisigned object from a dictionary of parameters.

        Args:
            value: The value to construct the SubmitMultisigned from.

        Returns:
            A new SubmitMultisigned object, constructed using the given parameters.
        """
        fixed_value = {**value}
        if "TransactionType" in fixed_value["tx_json"]:  # xrpl format
            fixed_value["tx_json"] = Transaction.from_xrpl(fixed_value["tx_json"])
        return super().from_dict(fixed_value)

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a SubmitMultisigned object.

        Returns:
            The dictionary representation of a SubmitMultisigned object.
        """
        xrpl_dict = super().to_dict()
        xrpl_dict["tx_json"] = self.tx_json.to_xrpl()
        return xrpl_dict



"""Enum containing the different Psuedo-Transaction types."""

from enum import Enum


class PseudoTransactionType(str, Enum):
    """Enum containing the different Psuedo-Transaction types."""

    ENABLE_AMENDMENT = "EnableAmendment"
    SET_FEE = "SetFee"
    UNL_MODIFY = "UNLModify"



"""
The server_info command asks the server for a
human-readable version of various information
about the rippled server being queried.
"""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class ServerDefinitions(Request):
    """
    The definitions command asks the server for a
    human-readable version of various information
    about the rippled server being queried.
    """

    method: RequestMethod = field(default=RequestMethod.SERVER_DEFINITIONS, init=False)

    hash: Optional[str] = None



"""The information needed to control an XRPL account."""

from __future__ import annotations

from typing import List, Optional, Type

from typing_extensions import Self

from xrpl.constants import CryptoAlgorithm, XRPLException
from xrpl.core import addresscodec
from xrpl.core.addresscodec import classic_address_to_xaddress, ensure_classic_address
from xrpl.core.addresscodec.exceptions import XRPLAddressCodecException
from xrpl.core.keypairs import derive_classic_address, derive_keypair, generate_seed


class Wallet:
    """
    The cryptographic keys needed to control an XRP Ledger account. See
    `Cryptographic Keys <https://xrpl.org/cryptographic-keys.html>`_ for
    details.
    """

    @property
    def address(self: Self) -> str:
        """
        The XRPL address that publicly identifies this wallet,
        as a base58 string. This is the same value as the `classic_address`.
        """  # noqa: DAR201
        return self._address

    # TODO: Just alias classic_address once mypy has resolved this issue:
    #       https://github.com/python/mypy/issues/6700
    @property
    def classic_address(self: Self) -> str:
        """
        `classic_address` is the same as `address`. It is called `classic_address` to
        differentiate it from the x-address standard, which encodes the network,
        destination tag, and XRPL address into a single value.
        It's also a base58 string.
        """  # noqa: DAR201
        return self._address

    def __init__(
        self: Self,
        public_key: str,
        private_key: str,
        *,
        master_address: Optional[str] = None,
        seed: Optional[str] = None,
        algorithm: Optional[CryptoAlgorithm] = None,
    ) -> None:
        """
        Generate a new Wallet.

        Args:
            public_key: The public key for the account.
            private_key: The private key used for signing transactions for the account.
            master_address: Include if a Wallet uses a Regular Key Pair. This sets the
                address that this wallet corresponds to. The default is `None`.
            seed: The seed used to derive the account keys. The default is `None`.
            algorithm: The algorithm used to encode the keys. Inferred from the seed if
                not included
        """
        if algorithm is None:
            if seed is not None and seed.startswith("sEd"):
                wallet_algorithm = CryptoAlgorithm.ED25519
            else:
                wallet_algorithm = CryptoAlgorithm.SECP256K1
        else:
            wallet_algorithm = algorithm

        """
        The core value that is used to derive all other information about
        this wallet. MUST be kept secret!
        """
        # Validate the seed before initialization of Wallet object
        if seed is not None:
            try:
                # We do not use the outputs of the below function, we are interested in
                # whether it throws an exception or not
                addresscodec.decode_seed(seed, wallet_algorithm)
            except Exception as e:
                raise XRPLAddressCodecException(
                    "Attempted to initialize a Wallet with "
                    + "an invalid seed: "
                    + seed
                    + ". The cryptographic algorithm"
                    + " used is: "
                    + wallet_algorithm
                    + "\nError message: "
                    + str(e)
                )

        self.seed = seed

        self.algorithm = wallet_algorithm
        """
        The algorithm that is used to convert the seed into its public/private keypair.
        """

        self.public_key = public_key
        """
        The public key that is used to identify this wallet's signatures, as
        a hexadecimal string.
        """

        self.private_key = private_key
        """
        The private key that is used to create signatures, as a hexadecimal
        string. MUST be kept secret!
        """

        self._address = (
            ensure_classic_address(master_address)
            if master_address is not None
            else derive_classic_address(self.public_key)
        )
        """Internal variable for classic_address. Use classic_address instead."""

    @classmethod
    def create(
        cls: Type[Self], algorithm: CryptoAlgorithm = CryptoAlgorithm.ED25519
    ) -> Self:
        """
        Generates a new seed and Wallet.

        Args:
            algorithm: The key-generation algorithm to use when generating the seed.
                The default is `ED25519`.

        Returns:
            The wallet that is generated from the given seed.
        """
        seed = generate_seed(algorithm=algorithm)
        return cls.from_seed(seed, algorithm=algorithm)

    @classmethod
    def from_seed(
        cls: Type[Self],
        seed: str,
        *,
        master_address: Optional[str] = None,
        algorithm: CryptoAlgorithm = CryptoAlgorithm.ED25519,
    ) -> Self:
        """
        Generates a new Wallet from seed (secret).

        Args:
            seed: The seed (secret) used to derive the account keys.
            master_address: Include if a Wallet uses a Regular Key Pair. This sets the
                address that this wallet corresponds to. The default is `None`.
            algorithm: The key-generation algorithm to use when generating the seed.
                The default is `ED25519`.

        Returns:
            The wallet that is generated from the given secret.
        """
        public_key, private_key = derive_keypair(seed, algorithm=algorithm)
        return cls(
            public_key,
            private_key,
            master_address=master_address,
            seed=seed,
            algorithm=algorithm,
        )

    from_secret = from_seed

    @classmethod
    def from_entropy(
        cls: Type[Self],
        entropy: str,
        *,
        master_address: Optional[str] = None,
        algorithm: CryptoAlgorithm = CryptoAlgorithm.ED25519,
    ) -> Self:
        """
        Generates a new Wallet from entropy (hexadecimal string of random numbers).

        Args:
            entropy: A hexadecimal string of random numbers to generate a seed used
                to derive a wallet.
            master_address: Include if a Wallet uses a Regular Key Pair. This sets the
                address that this wallet corresponds to. The default is `None`.
            algorithm: The key-generation algorithm to use when generating the seed.
                The default is `ED25519`.

        Returns:
            The wallet that is generated from the given entropy.

        Raises:
            XRPLException: If passed in entropy is not a bytestring.
        """
        if entropy is not None and len(entropy) != 32:
            raise XRPLException(
                "Entropy must be a 16-byte hexadecimal string of random numbers."
            )

        seed = generate_seed(entropy, algorithm)
        return cls.from_seed(seed, master_address=master_address, algorithm=algorithm)

    @classmethod
    def from_secret_numbers(
        cls: Type[Self],
        secret_numbers: List[str] | str,
        *,
        master_address: Optional[str] = None,
        algorithm: CryptoAlgorithm = CryptoAlgorithm.SECP256K1,
    ) -> Self:
        """
        Generates a new Wallet from secret numbers.

        Args:
            secret_numbers: A string (whitespace delimited) or string array consisting
                of 8 times 6 numbers used to derive a wallet.
            master_address: Include if a Wallet uses a Regular Key Pair. It must be
                the master address of the account. The default is `None`.
            algorithm: The digital signature algorithm to generate an address for.
                The default is `SECP256K1
                <https://xrpl.org/cryptographic-keys.html#secp256k1-key-derivation>`_
                (XUMM standard as of December 2022).

        Returns:
            The wallet that is generated from the given secret numbers.

        Raises:
            XRPLException: If the number of secret numbers is not 8. If the length of
                any secret number is not 6. If the checksum of any secret number is
                invalid.
        """
        # Logic adapted from xrpl-secret-numbers secretToEntropy function
        # https://github.com/XRPLF/xrpl.js/blob/main/packages/secret-numbers/src/utils/index.ts

        parsed_secret_numbers = (
            secret_numbers.split()
            if isinstance(secret_numbers, str)
            else secret_numbers
        )

        if len(parsed_secret_numbers) != 8:
            raise XRPLException("There must be 8 secret numbers.")

        entropy = ""
        for i, secret_number in enumerate(parsed_secret_numbers):
            no = int(secret_number[:5])
            checksum = int(secret_number[5:])

            if len(secret_number) != 6:
                raise XRPLException("Each secret number must be 6 digits long.")
            if no * (i * 2 + 1) % 9 != checksum:
                raise XRPLException(f"Checksum of secret number {i} is invalid.")

            hexed = hex(no)[2:].zfill(4)
            entropy += hexed

        return cls.from_entropy(
            entropy, master_address=master_address, algorithm=algorithm
        )

    def get_xaddress(
        self: Self, *, tag: Optional[int] = None, is_test: bool = False
    ) -> str:
        """
        Returns the X-Address of the Wallet's account.

        Args:
            tag: The destination tag of the address. Defaults to `None`.
            is_test: Whether the address corresponds to an address on the test network.
                Defaults to `False`.

        Returns:
            The X-Address of the Wallet's account.
        """
        return classic_address_to_xaddress(self.address, tag, is_test)

    def __str__(self: Self) -> str:
        """
        Returns a string representation of a Wallet.

        Returns:
            A string representation of a Wallet.
        """
        return "\n".join(
            [
                f"public_key: {self.public_key}",
                "private_key: -HIDDEN-",
                f"classic_address: {self.address}",
            ]
        )



"""
The submit method applies a transaction and sends it to the network to be confirmed and
included in future ledgers.

This command has two modes:
* Submit-only mode takes a signed, serialized transaction as a binary blob, and submits
it to the network as-is. Since signed transaction objects are immutable, no part of the
transaction can be modified or automatically filled in after submission.
* Sign-and-submit mode takes a JSON-formatted Transaction object, completes and signs
the transaction in the same manner as the sign method, and then submits the signed
transaction. We recommend only using this mode for testing and development.

To send a transaction as robustly as possible, you should construct and sign it in
advance, persist it somewhere that you can access even after a power outage, then
submit it as a tx_blob. After submission, monitor the network with the tx method
command to see if the transaction was successfully applied; if a restart or other
problem occurs, you can safely re-submit the tx_blob transaction: it won't be applied
twice since it has the same sequence number as the old transaction.

`See submit <https://xrpl.org/submit.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type

from typing_extensions import Self

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Submit(Request):
    """
    WARNING: This object should never be created. You should create an object of type
    `SignAndSubmit` or `SubmitOnly` instead.

    The submit method applies a transaction and sends it to the network to be confirmed
    and included in future ledgers.

    This command has two modes:
    * Submit-only mode takes a signed, serialized transaction as a binary blob, and
    submits it to the network as-is. Since signed transaction objects are immutable, no
    part of the transaction can be modified or automatically filled in after submission.
    * Sign-and-submit mode takes a JSON-formatted Transaction object, completes and
    signs the transaction in the same manner as the sign method, and then submits the
    signed transaction. We recommend only using this mode for testing and development.

    To send a transaction as robustly as possible, you should construct and sign it in
    advance, persist it somewhere that you can access even after a power outage, then
    submit it as a tx_blob. After submission, monitor the network with the tx method
    command to see if the transaction was successfully applied; if a restart or other
    problem occurs, you can safely re-submit the tx_blob transaction: it won't be
    applied twice since it has the same sequence number as the old transaction.

    `See submit <https://xrpl.org/submit.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SUBMIT, init=False)

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new Submit from a dictionary of parameters.

        Args:
            value: The value to construct the Submit from.

        Returns:
            A new Submit object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        from xrpl.models.requests import SignAndSubmit, SubmitOnly

        if cls.__name__ == "Submit":
            if "tx_blob" in value:
                return SubmitOnly.from_dict(value)  # type: ignore
            return SignAndSubmit.from_dict(value)  # type: ignore
        return super(Submit, cls).from_dict(value)



"""Async methods for working with XRPL wallets."""

from xrpl.asyncio.wallet.wallet_generation import (
    XRPLFaucetException,
    generate_faucet_wallet,
)

__all__ = ["XRPLFaucetException", "generate_faucet_wallet"]



"""Model for MPTokenAuthorize transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class MPTokenAuthorizeFlag(int, Enum):
    """
    Transactions of the MPTokenAuthorize type support additional values in the
    Flags field.
    This enum represents those options.
    """

    TF_MPT_UNAUTHORIZE = 0x00000001
    """
    If set and transaction is submitted by a holder, it indicates that the holder no
    longer wants to hold the MPToken, which will be deleted as a result. If the holder's
    MPToken has non-zero balance while trying to set this flag, the transaction will
    fail. On the other hand, if set and transaction is submitted by an issuer, it would
    mean that the issuer wants to unauthorize the holder (only applicable for
    allow-listing), which would unset the lsfMPTAuthorized flag on the MPToken.
    """


class MPTokenAuthorizeFlagInterface(TransactionFlagInterface):
    """
    Transactions of the MPTokenAuthorize type support additional values in the
    Flags field.
    This TypedDict represents those options.
    """

    TF_MPT_UNAUTHORIZE: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class MPTokenAuthorize(Transaction):
    """
    The MPTokenAuthorize transaction is used to globally lock/unlock a MPTokenIssuance,
    or lock/unlock an individual's MPToken.
    """

    mptoken_issuance_id: str = REQUIRED  # type: ignore
    """Identifies the MPTokenIssuance"""

    holder: Optional[str] = None
    """
    An optional XRPL Address of an individual token holder balance to lock/unlock.
    If omitted, this transaction will apply to all any accounts holding MPTs.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.MPTOKEN_AUTHORIZE,
        init=False,
    )



"""Model for DIDDelete transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field

from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class DIDDelete(Transaction):
    """Represents a DIDDelete transaction."""

    transaction_type: TransactionType = field(
        default=TransactionType.DID_DELETE,
        init=False,
    )



"""Model for DIDSet transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import HEX_REGEX, KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class DIDSet(Transaction):
    """Represents a DIDSet transaction."""

    did_document: Optional[str] = None
    """
    The DID document associated with the DID.

    To delete the Data, DIDDocument, or URI field from an existing DID ledger
    entry, add the field as an empty string.
    """

    data: Optional[str] = None
    """
    The public attestations of identity credentials associated with the DID.
    To delete the Data, DIDDocument, or URI field from an existing DID ledger
    entry, add the field as an empty string.
    """

    uri: Optional[str] = None
    """
    The Universal Resource Identifier associated with the DID.
    To delete the Data, DIDDocument, or URI field from an existing DID ledger
    entry, add the field as an empty string.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.DID_SET,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.did_document is None and self.data is None and self.uri is None:
            errors["did_set"] = "Must have one of `did_document`, `data`, and `uri`."
            # Can return here because there are no fields to process
            return errors

        if self.did_document == "" and self.data == "" and self.uri == "":
            errors["did_set"] = (
                "At least one of the fields `did_document`, `data`, and `uri` "
                + "must have a length greater than zero"
            )

            return errors

        def _process_field(name: str, value: Optional[str]) -> None:
            if value is not None:
                error_strs = []
                if not bool(HEX_REGEX.fullmatch(value)):
                    error_strs.append("must be hex")
                if len(value) > 256:
                    error_strs.append("must be <= 256 characters")
                if len(error_strs) > 0:
                    errors[name] = (" and ".join(error_strs) + ".").capitalize()

        _process_field("did_document", self.did_document)
        _process_field("data", self.data)
        _process_field("uri", self.uri)

        return errors



"""Model for a XChainModifyBridge transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.currencies import XRP
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


class XChainModifyBridgeFlag(int, Enum):
    """
    Transactions of the XChainModifyBridge type support additional values in the Flags
    field. This enum represents those options.
    """

    TF_CLEAR_ACCOUNT_CREATE_AMOUNT = 0x00010000


class XChainModifyBridgeFlagInterface(TransactionFlagInterface):
    """
    Transactions of the XChainModifyBridge type support additional values in the Flags
    field. This TypedDict represents those options.
    """

    TF_CLEAR_ACCOUNT_CREATE_AMOUNT: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XChainModifyBridge(Transaction):
    """
    Represents a XChainModifyBridge transaction.
    The XChainModifyBridge transaction allows bridge managers to modify the
    parameters of the bridge.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to modify. This field is required.

    :meta hide-value:
    """

    signature_reward: Optional[str] = None
    """
    The signature reward split between the witnesses for submitting
    attestations.

    :meta hide-value:
    """

    min_account_create_amount: Optional[str] = None
    """
    The minimum amount, in XRP, required for a ``XChainAccountCreateCommit``
    transaction. If this is not present, the ``XChainAccountCreateCommit``
    transaction will fail. This field can only be present on XRP-XRP bridges.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_MODIFY_BRIDGE,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge

        if (
            self.signature_reward is None
            and self.min_account_create_amount is None
            and not self.has_flag(XChainModifyBridgeFlag.TF_CLEAR_ACCOUNT_CREATE_AMOUNT)
        ):
            errors["xchain_modify_bridge"] = (
                "Must either change signature_reward, change "
                + "min_account_create_amount, or clear min_account_create_amount."
            )

        if self.account not in [bridge.locking_chain_door, bridge.issuing_chain_door]:
            errors["account"] = (
                "account must be either locking chain door or issuing chain door."
            )

        if self.signature_reward is not None and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "`signature_reward` must be numeric."

        if (
            self.min_account_create_amount is not None
            and bridge.locking_chain_issue != XRP()
        ):
            errors["min_account_create_amount"] = (
                "Cannot have MinAccountCreateAmount if bridge is IOU-IOU."
            )

        if (
            self.min_account_create_amount is not None
            and not self.min_account_create_amount.isnumeric()
        ):
            errors["min_account_create_amount_value"] = (
                "`min_account_create_amount` must be numeric."
            )

        return errors



"""
Specifies an amount in an issued currency, but without a value field.
This format is used for some book order requests.

See https://xrpl.org/currency-formats.html#specifying-currency-amounts
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Union

from typing_extensions import Self

import xrpl.models.amounts  # not a direct import, to get around circular imports
from xrpl.constants import HEX_MPTID_REGEX
from xrpl.models.base_model import BaseModel
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


def _is_valid_mptid(candidate: str) -> bool:
    return bool(HEX_MPTID_REGEX.fullmatch(candidate))


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class MPTCurrency(BaseModel):
    """
    Specifies an amount in an MPT, but without a value field.
    This format is used for some book order requests.

    See https://xrpl.org/currency-formats.html#specifying-currency-amounts
    """

    mpt_issuance_id: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if not _is_valid_mptid(self.mpt_issuance_id):
            errors["mpt_issuance_id"] = (
                f"Invalid mpt_issuance_id {self.mpt_issuance_id}"
            )
        return errors

    def to_amount(self: Self, value: Union[str, int]) -> xrpl.models.amounts.MPTAmount:
        """
        Converts an MPTCurrency to an MPTAmount.

        Args:
            value: The amount of MPTs in the MPTAmount.

        Returns:
            An MPTAmount that represents the MPT and the provided value.
        """
        return xrpl.models.amounts.MPTAmount(
            mpt_issuance_id=self.mpt_issuance_id, value=str(value)
        )



"""
Low-level functions for creating and using cryptographic keys with the XRP
Ledger.
"""

from xrpl.core.keypairs.exceptions import XRPLKeypairsException
from xrpl.core.keypairs.main import (
    derive_classic_address,
    derive_keypair,
    generate_seed,
    is_valid_message,
    sign,
)

__all__ = [
    "derive_classic_address",
    "derive_keypair",
    "generate_seed",
    "is_valid_message",
    "sign",
    "XRPLKeypairsException",
]
